import asyncio
import json
import math
import os
import shutil
import string
import random
import tempfile
import uuid
from typing import Optional

import aiofiles
import pytest

import sirius_sdk
from sirius_sdk.errors.indy_exceptions import WalletItemAlreadyExists
from sirius_sdk.encryption import create_keypair, bytes_to_b58
from sirius_sdk.agent.aries_rfc.feature_0750_storage import *
from sirius_sdk.agent.aries_rfc.feature_0750_storage.messages import BaseConfidentialStorageMessage, StreamOperation
from sirius_sdk.agent.aries_rfc.feature_0750_storage.errors import *
from sirius_sdk.agent.aries_rfc.feature_0750_storage.state_machines import CallerEncryptedDataVault, CalledEncryptedDataVault
from sirius_sdk.recipes.confidential_storage import SimpleDataVault

from tests.helpers import calc_file_hash, calc_bytes_hash, calc_file_size, run_coroutines
from tests.conftest import get_pairwise3


def get_random_string(length):
    # choose from all lowercase letter
    letters = string.ascii_lowercase
    s = ''.join(random.choice(letters) for i in range(length))
    return s


class MockedReadOnlyStream(FileSystemReadOnlyStream):
    """Mocked stream to test retentions and other stream behaviour"""

    def __init__(self, path: str, chunks_num: int, enc: Optional[StreamDecryption] = None):
        super().__init__(path, chunks_num, enc)
        # Set if need to emulate read-ops delays
        self.retention_delay_sec: Optional[int] = None
        # Set to limit retention factor
        self.retention_delay_limit: Optional[int] = None
        self.__retention_delay_accum = 0

    async def read_chunk(self, no: int = None) -> (int, bytes):
        if self.retention_delay_sec:
            self.__retention_delay_accum += 1
            if self.retention_delay_limit is not None:
                if self.__retention_delay_accum >= self.retention_delay_limit:
                    allow_sleeping = False
                    self.__retention_delay_accum = 0
                else:
                    allow_sleeping = True
            else:
                allow_sleeping = True
            if allow_sleeping:
                await asyncio.sleep(self.retention_delay_sec)
        return await super().read_chunk(no)


@pytest.mark.asyncio
async def test_fs_streams(files_dir: str):
    file_under_test = os.path.join(files_dir, 'big_img.jpeg')
    file_under_test_md5 = calc_file_hash(file_under_test)
    file_under_test_size = calc_file_size(file_under_test)
    # 1. Reading All Data
    for chunks_num in [10, 5, 1]:
        ro = FileSystemReadOnlyStream(file_under_test, chunks_num)
        await ro.open()
        try:
            read_data = await ro.read()
            read_data_md5 = calc_bytes_hash(read_data)
            assert read_data_md5 == file_under_test_md5
            assert ro.current_chunk == chunks_num
            assert await ro.eof() is True
        finally:
            await ro.close()
    # 2. Reading with chunks
    for chunks_num in [10, 5, 1]:
        ro = FileSystemReadOnlyStream(file_under_test, chunks_num)
        await ro.open()
        try:
            accum = b''
            async for chunk in ro.read_chunked():
                accum += chunk
            accum_md5 = calc_bytes_hash(read_data)
            assert accum_md5 == file_under_test_md5
            assert ro.current_chunk == chunks_num
            assert await ro.eof() is True
        finally:
            await ro.close()
    # 3. Write stream: create and write all data to file
    wo_file_path = os.path.join(files_dir, 'big_img_writeonly_stream.jpeg')
    for chunk_size in [100, 1024, 10000000]:
        wo = FileSystemWriteOnlyStream(wo_file_path, chunk_size)
        expected_chunks_num = math.ceil(file_under_test_size/chunk_size)
        await wo.create(truncate=True)
        try:
            with open(file_under_test, 'rb') as f:
                raw = f.read()
            await wo.open()
            try:
                assert wo.is_open is True
                await wo.write(raw)
                assert wo.chunks_num == expected_chunks_num
                assert wo.current_chunk == expected_chunks_num
            finally:
                await wo.close()
            wo_file_md5 = calc_file_hash(wo_file_path)
            assert wo_file_md5 == file_under_test_md5
        finally:
            os.remove(wo_file_path)
    # 4. Write stream: append to end
    wo_file_path = os.path.join(files_dir, 'append_stream.jpeg')
    chunk_size = 1
    wo = FileSystemWriteOnlyStream(wo_file_path, chunk_size)
    await wo.create(truncate=True)
    await wo.open()
    try:
        await wo.write(b'0' * chunk_size)
        assert wo.current_chunk == 1
        await wo.close()
        wo = FileSystemWriteOnlyStream(wo_file_path, chunk_size)
        await wo.open()
        assert wo.current_chunk == 1
        await wo.write(b'1' * chunk_size)
        await wo.close()
        with open(wo_file_path, 'rb') as f:
            content = f.read()
            assert content == b'01'
    finally:
        os.remove(wo_file_path)


@pytest.mark.asyncio
async def test_fs_streams_seeks(files_dir: str):
    file_under_test = os.path.join(files_dir, 'seeks.bin')
    chunks = [b'chunk1', b'chunk2', b'chunk3']
    assert len(chunks[0]) == len(chunks[1]) == len(chunks[2])
    wo = FileSystemWriteOnlyStream(file_under_test, chunk_size=len(chunks[0]))
    await wo.create(truncate=True)
    try:
        await wo.open()
        try:
            # Check Writer
            assert wo.chunks_num == 0
            for no, chunk in enumerate(chunks):
                chunk_pos, writen = await wo.write_chunk(chunk, no)
                assert chunk_pos == no+1
                assert writen == len(chunk)
            assert wo.chunks_num == len(chunks)
            # Check Writer EOF
            with pytest.raises(StreamEOF):
                await wo.seek_to_chunk(1000)
            with pytest.raises(StreamEOF):
                await wo.write_chunk(b'', 1000)
            # Check Reader
            ro = FileSystemReadOnlyStream(file_under_test, chunks_num=len(chunks))
            await ro.open()
            try:
                assert ro.is_open is True
                assert ro.current_chunk == 0
                assert ro.chunks_num == len(chunks)
                for no, expected_chunk in enumerate(chunks):
                    await ro.seek_to_chunk(no)
                    chunk_offset, actual_chunk = await ro.read_chunk(no)
                    assert chunk_offset == no+1
                    assert expected_chunk == actual_chunk
                # Check Reader EOF
                assert await ro.eof() is True
                with pytest.raises(StreamEOF):
                    await ro.seek_to_chunk(1000)
                with pytest.raises(StreamEOF):
                    await ro.read_chunk()
            finally:
                await ro.close()
        finally:
            await wo.close()
    finally:
        os.remove(file_under_test)


@pytest.mark.asyncio
async def test_fs_streams_truncate():
    chunks = [b'chunk1', b'chunk2', b'chunk3']
    actual_chunks_num = len(chunks)
    for trunked_to_no in [0, 1, 2, 1000]:
        file_under_test = os.path.join(tempfile.tempdir, f'truncates_for_{trunked_to_no}.bin')
        wo = FileSystemWriteOnlyStream(file_under_test, chunk_size=len(chunks[0]))
        await wo.create(truncate=True)
        try:
            await wo.open()
            try:
                assert wo.chunks_num == 0
                for no, chunk in enumerate(chunks):
                    await wo.write_chunk(chunk, no)
                assert wo.chunks_num == len(chunks)
                await wo.truncate(trunked_to_no)
                assert wo.chunks_num == min(trunked_to_no, actual_chunks_num), f'Error for trancate to {trunked_to_no}'
                assert wo.current_chunk == min(trunked_to_no, actual_chunks_num), f'Error for trancate to {trunked_to_no}'
            finally:
                await wo.close()
            with open(file_under_test, 'rb') as f:
                content = f.read()
            expected = ''.join([s.decode() for s in chunks[:trunked_to_no]]).encode()
            assert expected == content, f'Error for trancate to {trunked_to_no}'
        finally:
            os.remove(file_under_test)


@pytest.mark.asyncio
async def test_fs_streams_copy(files_dir: str):
    file_under_test = os.path.join(files_dir, 'big_img.jpeg')
    file_under_test_md5 = calc_file_hash(file_under_test)
    file_under_test_size = calc_file_size(file_under_test)
    chunks_num = 5
    # 1. Reading All Data
    for chunk_size in [100, 1024, 10000000]:
        expected_chunks_num = chunks_num  # math.ceil(file_under_test_size / chunk_size)
        ro = FileSystemReadOnlyStream(file_under_test, chunks_num=chunks_num)
        await ro.open()
        try:
            wo_file_path = os.path.join(files_dir, 'big_img_copy_stream.jpeg')
            try:
                wo = FileSystemWriteOnlyStream(wo_file_path, chunk_size)
                await wo.create(truncate=True)
                await wo.open()
                try:
                    await wo.copy(ro)
                finally:
                    await wo.close()
                with open(wo_file_path, 'rb') as f:
                    raw = f.read()
                copied_md5 = calc_bytes_hash(raw)
                assert copied_md5 == file_under_test_md5
                assert await ro.eof() is True
                assert wo.chunks_num == expected_chunks_num
            finally:
                os.remove(wo_file_path)
        finally:
            await ro.close()


@pytest.mark.asyncio
async def test_fs_streams_encoding(files_dir: str):
    file_under_test = os.path.join(files_dir, 'big_img.jpeg')
    file_under_test_md5 = calc_file_hash(file_under_test)

    recip_vk_bytes, recip_sigkey_bytes = create_keypair(b'00000000000000000000000RECIPIENT')
    recip_vk, recip_sigkey = bytes_to_b58(recip_vk_bytes), bytes_to_b58(recip_sigkey_bytes)

    write_enc = StreamEncryption()
    write_enc.setup(target_verkeys=[recip_vk])
    with open(file_under_test, 'rb') as f:
        file_content = f.read()
    enc_file_path = os.path.join(files_dir, 'big_img_encrypted_stream.jpeg.bin')
    wo = FileSystemWriteOnlyStream(enc_file_path, chunk_size=1024*2, enc=write_enc)
    try:
        await wo.create(truncate=True)
        await wo.open()
        try:
            await wo.write(file_content)
            chunks_num = wo.chunks_num
        finally:
            await wo.close()
        read_enc = StreamDecryption(recipients=write_enc.recipients, nonce=write_enc.nonce)
        read_enc.setup(recip_vk, recip_sigkey)
        ro = FileSystemReadOnlyStream(enc_file_path, chunks_num=chunks_num, enc=read_enc)
        await ro.open()
        try:
            data = await ro.read()
            data_md5 = calc_bytes_hash(data)
            assert data_md5 == file_under_test_md5
        finally:
            await ro.close()
    finally:
        os.remove(enc_file_path)


@pytest.mark.asyncio
async def test_fs_streams_encoding_with_jwe(files_dir: str):
    file_under_test = os.path.join(files_dir, 'big_img.jpeg')
    file_under_test_md5 = calc_file_hash(file_under_test)

    recip_vk_bytes1, recip_sigkey_bytes1 = create_keypair(b'0000000000000000000000RECIPIENT1')
    recip_vk1, recip_sigkey1 = bytes_to_b58(recip_vk_bytes1), bytes_to_b58(recip_sigkey_bytes1)
    recip_vk_bytes2, recip_sigkey_bytes2 = create_keypair(b'0000000000000000000000RECIPIENT2')
    recip_vk2, recip_sigkey2 = bytes_to_b58(recip_vk_bytes2), bytes_to_b58(recip_sigkey_bytes2)

    write_enc = StreamEncryption()
    write_enc.setup(target_verkeys=[recip_vk1, recip_vk2])
    jwe = write_enc.jwe
    print('')

    with open(file_under_test, 'rb') as f:
        file_content = f.read()
    enc_file_path = os.path.join(files_dir, 'big_img_encrypted_stream.jpeg.bin')
    wo = FileSystemWriteOnlyStream(enc_file_path, chunk_size=1024*2, enc=write_enc)
    try:
        await wo.create(truncate=True)
        await wo.open()
        try:
            await wo.write(file_content)
            chunks_num = wo.chunks_num
        finally:
            await wo.close()

        for recip_vk, recip_sigkey in [(recip_vk1, recip_sigkey1), (recip_vk2, recip_sigkey2)]:
            read_enc = StreamDecryption()
            read_enc.jwe = jwe
            read_enc.setup(recip_vk, recip_sigkey)

            ro = FileSystemReadOnlyStream(enc_file_path, chunks_num=chunks_num, enc=read_enc)
            await ro.open()
            try:
                data = await ro.read()
                data_md5 = calc_bytes_hash(data)
                assert data_md5 == file_under_test_md5
            finally:
                await ro.close()
    finally:
        os.remove(enc_file_path)


@pytest.mark.asyncio
async def test_fs_storage(files_dir: str, config_a: dict):
    uri_under_test = 'document.bin'
    test_data = b'Test-Data'
    path = os.path.join(files_dir, uri_under_test)
    stored_jwe = None
    stored_cek = None
    if os.path.isfile(path):
        os.remove(path)
    try:
        async with sirius_sdk.context(**config_a):
            vk = await sirius_sdk.Crypto.create_key()
            enc1 = StreamEncryption()
            enc1.setup(target_verkeys=[vk])
            stored_jwe = enc1.jwe
            stored_cek = enc1.cek
            storage = FileSystemRawByteStorage(enc1)
            await storage.mount(files_dir)
            await storage.create(uri_under_test)
            wo = await storage.writeable(uri_under_test)
            await wo.open()
            await wo.write(test_data)
            await wo.close()
            ro = await storage.readable(uri_under_test, chunks_num=wo.chunks_num)
            await ro.open()
            try:
                actual_data = await ro.read()
            except Exception as e:
                raise e
            await ro.close()
            assert actual_data == test_data
            # RE-Create storage for same JWE
            enc2 = StreamEncryption.from_jwe(stored_jwe, stored_cek)
            storage = FileSystemRawByteStorage(enc2)
            await storage.mount(files_dir)
            ro = await storage.readable(uri_under_test, chunks_num=wo.chunks_num)
            await ro.open()
            actual_data = await ro.read()
            await ro.close()
            assert actual_data == test_data
            wo = await storage.writeable(uri_under_test)
            await wo.open()
            await wo.truncate()
            try:
                await wo.write(test_data)
            except Exception as e:
                raise e
            await wo.close()
            ro = await storage.readable(uri_under_test, chunks_num=wo.chunks_num)
            await ro.open()
            try:
                actual_data = await ro.read()
            except Exception as e:
                raise
            await ro.close()
            assert actual_data == test_data
    finally:
        if os.path.isfile(path):
            os.remove(path)


@pytest.mark.asyncio
async def test_fs_streams_truncate_for_encoding():
    chunks = [b'chunk1', b'chunk2', b'chunk3']
    actual_chunks_num = len(chunks)
    for trunked_to_no in [0, 1, 2, 1000]:
        file_under_test = os.path.join(tempfile.tempdir, f'truncates_for_{trunked_to_no}.bin')
        wo = FileSystemWriteOnlyStream(
            file_under_test, chunk_size=len(chunks[0]), enc=StreamEncryption(type_=ConfidentialStorageEncType.UNKNOWN)
        )
        await wo.create(truncate=True)
        try:
            await wo.open()
            try:
                assert wo.chunks_num == 0
                for no, chunk in enumerate(chunks):
                    await wo.write_chunk(chunk, no)
                assert wo.chunks_num == len(chunks)
                await wo.truncate(trunked_to_no)
                assert wo.chunks_num == min(trunked_to_no, actual_chunks_num), f'Error for truncate to {trunked_to_no}'
                assert wo.current_chunk == min(trunked_to_no, actual_chunks_num), f'Error for truncate to {trunked_to_no}'
            finally:
                await wo.close()
            if trunked_to_no == 0:
                with open(file_under_test, 'rb') as f:
                    content = f.read()
                assert content == b''
            else:
                ro = FileSystemReadOnlyStream(
                    file_under_test, chunks_num=min(trunked_to_no, actual_chunks_num), enc=StreamDecryption(type_=ConfidentialStorageEncType.UNKNOWN)
                )
                await ro.open()
                try:
                    content = await ro.read()
                finally:
                    await ro.close()
                expected = ''.join([s.decode() for s in chunks[:trunked_to_no]]).encode()
                assert expected == content, f'Error for truncate to {trunked_to_no}'
        finally:
            os.remove(file_under_test)


@pytest.mark.asyncio
async def test_fs_streams_decoding_from_wallet1(config_c: dict):

    async with sirius_sdk.context(**config_c):
        seed = uuid.uuid4().hex[:32]
        vk = await sirius_sdk.Crypto.create_key(seed=seed)

    # Non-ASCII symbols
    chunk = "hello aåbäcö".encode()

    enc = StreamEncryption(nonce=bytes_to_b58(b'0'*12)).setup(target_verkeys=[vk])
    dec = StreamDecryption(recipients=enc.recipients, nonce=enc.nonce)

    # 1. Test enc/dec
    assert dec.cek is None

    async with sirius_sdk.context(**config_c):
        w = FileSystemWriteOnlyStream(path='', chunk_size=1, enc=enc)
        r = FileSystemReadOnlyStream(path='', chunks_num=1, enc=dec)
        encoded = await w.encrypt(chunk)
        decoded = await r.decrypt(encoded)
        assert decoded == chunk

    # 2. Test with JWE
    jwe = enc.jwe
    js = jwe.as_json()
    restored_jwe = JWE()
    restored_jwe.from_json(js)
    dec = StreamDecryption.from_jwe(restored_jwe)
    assert dec.cek is None
    async with sirius_sdk.context(**config_c):
        w = FileSystemWriteOnlyStream(path='', chunk_size=1, enc=enc)
        r = FileSystemReadOnlyStream(path='', chunks_num=1, enc=dec)
        encoded = await w.encrypt(chunk)
        decoded = await r.decrypt(encoded)
        assert decoded == chunk


@pytest.mark.asyncio
async def test_fs_streams_decoding_from_wallet2(config_c: dict):

    async with sirius_sdk.context(**config_c):
        vk = await sirius_sdk.Crypto.create_key()

    enc = StreamEncryption().setup(target_verkeys=[vk])
    file_under_test = os.path.join(tempfile.tempdir, f'{uuid.uuid4().hex}.bin')
    chunk_size = 1024
    chunk = b'x' * chunk_size

    wo = FileSystemWriteOnlyStream(file_under_test, chunk_size=chunk_size, enc=enc)
    await wo.create(truncate=True)
    try:
        await wo.open()
        await wo.write(chunk)
        test_file_chunks_num = wo.chunks_num
        await wo.close()
        # Decoding
        dec = StreamDecryption(type_=enc.type, recipients=enc.recipients, nonce=enc.nonce)
        async with sirius_sdk.context(**config_c):
            ro = FileSystemReadOnlyStream(file_under_test, chunks_num=test_file_chunks_num, enc=dec)
            await ro.open()
            no, actual_chunk = await ro.read_chunk()
            await ro.close()
            assert actual_chunk == chunk
    finally:
        os.remove(file_under_test)


@pytest.mark.asyncio
async def test_streams_encoding_layers(files_dir: str):
    file_under_test = os.path.join(files_dir, 'big_img.jpeg')
    file_under_test_md5 = calc_file_hash(file_under_test)
    tmp_file_layer1 = os.path.join(tempfile.tempdir, f'layer1_{uuid.uuid4().hex}.bin')
    tmp_file_layer2 = os.path.join(tempfile.tempdir, f'layer2_{uuid.uuid4().hex}.bin')
    file_for_checks = os.path.join(tempfile.tempdir, f'checks_{uuid.uuid4().hex}.bin')
    # Layer-1
    recip_vk_bytes1, recip_sigkey_bytes1 = create_keypair(b'0000000000000000000000RECIPIENT1')
    recip_vk1, recip_sigkey1 = bytes_to_b58(recip_vk_bytes1), bytes_to_b58(recip_sigkey_bytes1)
    layers_chunk_size = 1024
    layer1 = {
        'keys': (recip_vk1, recip_sigkey1),
        'writer': FileSystemWriteOnlyStream(
            path=tmp_file_layer1,
            chunk_size=layers_chunk_size,
            enc=StreamEncryption().setup(target_verkeys=[recip_vk1])
        )
    }
    # Layer-2
    recip_vk_bytes2, recip_sigkey_bytes2 = create_keypair(b'0000000000000000000000RECIPIENT2')
    recip_vk2, recip_sigkey2 = bytes_to_b58(recip_vk_bytes2), bytes_to_b58(recip_sigkey_bytes2)
    layer2 = {
        'keys': (recip_vk2, recip_sigkey2),
        'writer': FileSystemWriteOnlyStream(
            path=tmp_file_layer2,
            chunk_size=layers_chunk_size,
            enc=StreamEncryption().setup(target_verkeys=[recip_vk2])
        )
    }
    await layer1['writer'].create(truncate=True)
    try:
        await layer2['writer'].create(truncate=True)
        try:
            with open(file_under_test, 'rb') as f:
                content_under_test = f.read()
            # Layer-2 Encrypt (upper-level)
            await layer2['writer'].open()
            await layer2['writer'].write(content_under_test)
            layer2_chunks_num = layer2['writer'].chunks_num
            await layer2['writer'].close()
            # Layer-1 Encrypt (Lower-Level)
            src = FileSystemReadOnlyStream(
                path=layer2['writer'].path, chunks_num=layer2_chunks_num, enc=StreamDecryption(type_=ConfidentialStorageEncType.UNKNOWN)
            )
            await src.open()
            await layer1['writer'].create(truncate=True)
            await layer1['writer'].open()
            await layer1['writer'].copy(src)
            layer1_chunks_num = layer1['writer'].chunks_num
            await layer1['writer'].close()
            await src.close()
            # Layer1 -> Layer2 decryption (lower -> upper)
            layer1_reader = FileSystemReadOnlyStream(
                path=layer1['writer'].path, chunks_num=layer1_chunks_num, enc=layer1['writer'].enc
            )
            await layer1_reader.open()
            try:
                dest = FileSystemWriteOnlyStream(
                    path=file_for_checks, chunk_size=layers_chunk_size, enc=StreamEncryption(type_=ConfidentialStorageEncType.UNKNOWN)
                )
                await dest.create(truncate=True)
                try:
                    await dest.open()
                    await dest.copy(layer1_reader)
                    await dest.close()
                    layer2_reader = FileSystemReadOnlyStream(
                        path=file_for_checks, chunks_num=layer2_chunks_num, enc=layer2['writer'].enc
                    )
                    await layer2_reader.open()
                    try:
                        # Check-1
                        actual_data = await layer2_reader.read()
                        actual_md5 = calc_bytes_hash(actual_data)
                        assert file_under_test_md5 == actual_md5
                        # Check-2
                        accum = b''
                        async for chunk in layer2_reader.read_chunked(src=layer1_reader):
                            accum += chunk
                        assert file_under_test_md5 == actual_md5
                    finally:
                        await layer2_reader.close()
                finally:
                    os.remove(file_for_checks)
            finally:
                await layer1_reader.close()
        finally:
            os.remove(tmp_file_layer2)
    finally:
        os.remove(tmp_file_layer1)


@pytest.mark.asyncio
async def test_streams_encoding_seeking():
    chunk_size = 1024
    actual_chunks = [
        b'0' * chunk_size,
        b'1' * chunk_size,
        b'2' * (chunk_size//2)
    ]
    append_chunk = b'x' * chunk_size

    recip_vk_bytes, recip_sigkey_bytes = create_keypair(b'0000000000000000000000RECIPIENT1')
    recip_vk, recip_sigkey = bytes_to_b58(recip_vk_bytes), bytes_to_b58(recip_sigkey_bytes)
    file_under_test = os.path.join(tempfile.tempdir, f'{uuid.uuid4().hex}.bin')
    open(file_under_test, 'w+b')
    try:
        enc = StreamEncryption(
            type_=ConfidentialStorageEncType.X25519KeyAgreementKey2019).setup(target_verkeys=[recip_vk]
                                                                              )
        writer = FileSystemWriteOnlyStream(
            path=file_under_test, chunk_size=chunk_size,
            enc=enc
        )
        # Write chunks
        await writer.open()
        try:
            for no, chunk in enumerate(actual_chunks):
                await writer.write_chunk(chunk)
            assert writer.chunks_num == len(actual_chunks)
        finally:
            await writer.close()
        # Read chunks
        reader = FileSystemReadOnlyStream(
            path=file_under_test, chunks_num=len(actual_chunks),
            enc=StreamDecryption(
                recipients=writer.enc.recipients, type_=ConfidentialStorageEncType.X25519KeyAgreementKey2019, nonce=writer.enc.nonce
            ).setup(recip_vk, recip_sigkey)
        )
        await reader.open()
        try:
            for no, expected_chunk in enumerate(actual_chunks):
                new_no, actual_chunk = await reader.read_chunk(no)
                assert actual_chunk == expected_chunk
                assert new_no == no+1
        finally:
            await reader.close()
        # Append
        writer = FileSystemWriteOnlyStream(
            path=file_under_test, chunk_size=chunk_size,
            enc=enc
        )
        # Append chunk to the end of stream
        await writer.open()
        try:
            await writer.write_chunk(append_chunk)
            actual_chunks_after_append = writer.chunks_num
            assert actual_chunks_after_append == len(actual_chunks) + 1
        finally:
            await writer.close()
        reader = FileSystemReadOnlyStream(
            path=file_under_test, chunks_num=actual_chunks_after_append,
            enc=StreamDecryption(
                recipients=writer.enc.recipients, type_=ConfidentialStorageEncType.X25519KeyAgreementKey2019, nonce=writer.enc.nonce
            ).setup(recip_vk, recip_sigkey)
        )
        await reader.open()
        try:
            i = 0
            async for chunk in reader.read_chunked():
                if i < len(actual_chunks):
                    expected = actual_chunks[i]
                else:
                    expected = append_chunk
                assert chunk == expected
                i += 1
        finally:
            await reader.close()
    finally:
        os.remove(file_under_test)


@pytest.mark.asyncio
async def test_fs_streams_encoding_copy(files_dir: str):
    file_under_test = os.path.join(files_dir, 'big_img.jpeg')
    file_under_test_md5 = calc_file_hash(file_under_test)

    sender_vk_bytes, sender_sigkey_bytes = create_keypair(b'00000000000000000000000000SENDER')
    sender_vk, sender_sigkey = bytes_to_b58(sender_vk_bytes), bytes_to_b58(sender_sigkey_bytes)
    recip_vk_bytes, recip_sigkey_bytes = create_keypair(b'00000000000000000000000RECIPIENT')
    recip_vk, recip_sigkey = bytes_to_b58(recip_vk_bytes), bytes_to_b58(recip_sigkey_bytes)

    write_enc = StreamEncryption()
    write_enc.setup(target_verkeys=[recip_vk])
    with open(file_under_test, 'rb') as f:
        file_content = f.read()
    enc_file_path = os.path.join(files_dir, 'big_img_encrypted_stream.jpeg.bin')
    wo = FileSystemWriteOnlyStream(enc_file_path, chunk_size=1024, enc=write_enc)
    try:
        await wo.create(truncate=True)
        await wo.open()
        try:
            await wo.write(file_content)
        finally:
            await wo.close()
        read_enc = StreamDecryption(recipients=write_enc.recipients, nonce=write_enc.nonce)
        read_enc.setup(recip_vk, recip_sigkey)
        ro = FileSystemReadOnlyStream(enc_file_path, chunks_num=wo.chunks_num, enc=read_enc)
        await ro.open()
        try:
            enc_copy_path = os.path.join(files_dir, 'big_img_encrypted_stream_copy.jpeg')
            try:
                copy_wo = FileSystemWriteOnlyStream(enc_copy_path, chunk_size=1024)
                await copy_wo.create(truncate=True)
                await copy_wo.open()
                try:
                    await copy_wo.copy(ro)
                finally:
                    await copy_wo.close()
                copied_file_md5 = calc_file_hash(enc_copy_path)
                assert copied_file_md5 == file_under_test_md5
            finally:
                os.remove(enc_copy_path)
        finally:
            await ro.close()
    finally:
        os.remove(enc_file_path)


@pytest.mark.asyncio
async def test_streams_encoding_decoding_wrappers(files_dir: str, config_a: dict):
    # Layer-2
    target_vk, target_sk = create_keypair(seed='00000000000000000000000000TARGET'.encode())
    target_vk, target_sk = bytes_to_b58(target_vk), bytes_to_b58(target_sk)
    # Layer-1
    storage_vk, storage_sk = create_keypair(seed='0000000000000000000000000STORAGE'.encode())
    storage_vk, storage_sk = bytes_to_b58(storage_vk), bytes_to_b58(storage_sk)
    # File on Disk
    test_data = b'data*' * 1024
    storage_file = os.path.join(files_dir, 'big_img_encrypted_stream.jpeg.bin')
    # Create write-stream on Layer-1
    wo = FileSystemWriteOnlyStream(
        storage_file, chunk_size=1024, enc=StreamEncryption().setup(target_verkeys=[storage_vk])
    )
    jwe_layer1 = wo.enc.jwe
    await wo.create(truncate=True)
    try:
        # Wrap with stream abstraction on Layer-2
        enc = StreamEncryption().setup(target_verkeys=[target_vk])
        jwe = enc.jwe
        enc_restored = StreamEncryption.from_jwe(jwe)
        wrapper_to_write = WriteOnlyStreamEncodingWrapper(
            dest=wo, enc=enc_restored
        )
        jwe_layer2 = wrapper_to_write.enc.jwe
        await wrapper_to_write.open()
        try:
            await wrapper_to_write.write(test_data)
        finally:
            await wrapper_to_write.close()
        # Open Read stream on Layer-1
        ro = FileSystemReadOnlyStream(
            storage_file, chunks_num=wo.chunks_num,
            enc=StreamDecryption.from_jwe(jwe_layer1).setup(storage_vk, storage_sk)
        )
        # Wrap with stream abstraction on Layer-2
        wrapper_to_read = ReadOnlyStreamDecodingWrapper(
            src=ro,
            enc=StreamDecryption.from_jwe(jwe_layer2).setup(target_vk, target_sk)
        )
        await wrapper_to_read.open()
        try:
            actual_data = await wrapper_to_read.read()
            assert actual_data == test_data
        finally:
            await wrapper_to_read.close()
        # Wrap with stream on Layer-2 via sirius_sdk.Crypto
        async with sirius_sdk.context(**config_a):
            try:
                await sirius_sdk.Crypto.create_key(seed='00000000000000000000000000TARGET')
            except WalletItemAlreadyExists:
                pass
            wrapper_to_read = ReadOnlyStreamDecodingWrapper(
                src=ro,
                enc=StreamDecryption.from_jwe(jwe_layer2)
            )
            await wrapper_to_read.open()
            try:
                actual_data = await wrapper_to_read.read()
                assert actual_data == test_data
            finally:
                await wrapper_to_read.close()
    finally:
        os.remove(storage_file)


@pytest.mark.asyncio
async def test_readonly_stream_protocols(files_dir: str, config_c: dict, config_d: dict):
    """
    Agent-C is CALLER
    Agent-D is CALLED
    """

    caller_p2p = await get_pairwise3(me=config_c, their=config_d)
    called_p2p = await get_pairwise3(me=config_d, their=config_c)
    testing_thid = 'test-readonly-protocol-' + uuid.uuid4().hex
    file_under_test = os.path.join(files_dir, 'big_img.jpeg')
    ro_chunks_num = 10
    etalon_chunks = []
    print('#')

    etalon_ro = FileSystemReadOnlyStream(path=file_under_test, chunks_num=ro_chunks_num)
    await etalon_ro.open()
    try:
        async for chunk in etalon_ro.read_chunked():
            etalon_chunks.append(chunk)
    finally:
        await etalon_ro.close()

    async def called():
        async with sirius_sdk.context(**config_d):
            stream = FileSystemReadOnlyStream(path=file_under_test, chunks_num=ro_chunks_num)
            state_machine = CalledReadOnlyStreamProtocol(called_p2p, thid=testing_thid)
            await state_machine.run_forever(stream)

    async def caller():
        async with sirius_sdk.context(**config_c):
            ro = CallerReadOnlyStreamProtocol(called=caller_p2p, uri=file_under_test, read_timeout=5, thid=testing_thid)
            # open
            await ro.open()
            # checks...
            assert ro.is_open is True
            assert ro.current_chunk == 0
            assert ro.chunks_num == ro_chunks_num
            assert ro.seekable is True
            # read
            no1, chunk1 = await ro.read_chunk()
            assert ro.current_chunk == no1
            assert etalon_chunks[no1-1] == chunk1
            no2, chunk2 = await ro.read_chunk()
            assert ro.current_chunk == no2
            assert etalon_chunks[no2-1] == chunk2
            # seeks
            no = await ro.seek_to_chunk(0)
            assert no == 0
            assert ro.current_chunk == 0
            no = await ro.seek_to_chunk(3)
            assert no == 3
            assert ro.current_chunk == 3
            # close
            await ro.close()
            assert ro.is_open is False

    results = await run_coroutines(called(), caller(), timeout=5)
    assert len(results) == 2


@pytest.mark.asyncio
async def test_readonly_stream_protocols_chunks_and_eof(files_dir: str, config_c: dict, config_d: dict):
    """
    Agent-C is CALLER
    Agent-D is CALLED
    """

    caller_p2p = await get_pairwise3(me=config_c, their=config_d)
    called_p2p = await get_pairwise3(me=config_d, their=config_c)
    testing_thid = 'test-readonly-protocol-' + uuid.uuid4().hex
    file_under_test = os.path.join(files_dir, 'big_img.jpeg')
    file_under_test_md5 = calc_file_hash(file_under_test)
    ro_chunks_num = 10

    async def called():
        async with sirius_sdk.context(**config_d):
            stream = FileSystemReadOnlyStream(path=file_under_test, chunks_num=ro_chunks_num)
            state_machine = CalledReadOnlyStreamProtocol(called_p2p, thid=testing_thid)
            await state_machine.run_forever(stream)

    async def caller():
        async with sirius_sdk.context(**config_c):
            ro = CallerReadOnlyStreamProtocol(called=caller_p2p, uri=file_under_test, read_timeout=5, thid=testing_thid)
            # open
            await ro.open()
            # checks read data md5
            read_data = await ro.read()
            assert calc_bytes_hash(read_data) == file_under_test_md5
            # check EOF
            with pytest.raises(StreamEOF):
                await ro.seek_to_chunk(1000)
            await ro.seek_to_chunk(ro_chunks_num)
            with pytest.raises(StreamEOF):
                await ro.read_chunk()
            # close
            await ro.close()

    results = await run_coroutines(called(), caller(), timeout=5)
    assert len(results) == 2


@pytest.mark.asyncio
async def test_readonly_stream_delays(files_dir: str, config_c: dict, config_d: dict):
    """Check SWAP mechanism in non-reliable communication channel
    Agent-C is CALLER
    Agent-D is CALLED
    """
    caller_p2p = await get_pairwise3(me=config_c, their=config_d)
    called_p2p = await get_pairwise3(me=config_d, their=config_c)
    testing_thid = 'test-readonly-protocol-' + uuid.uuid4().hex
    file_under_test = os.path.join(files_dir, 'big_img.jpeg')
    file_under_test_md5 = calc_file_hash(file_under_test)
    ro_chunks_num = 10

    async def called():
        async with sirius_sdk.context(**config_d):
            stream = MockedReadOnlyStream(path=file_under_test, chunks_num=ro_chunks_num)
            stream.retention_delay_sec = 1
            stream.retention_delay_limit = 2
            state_machine = CalledReadOnlyStreamProtocol(called_p2p, thid=testing_thid)
            await state_machine.run_forever(stream)

    async def caller():
        async with sirius_sdk.context(**config_c):
            ro = CallerReadOnlyStreamProtocol(
                called=caller_p2p, uri=file_under_test, read_timeout=3, retry_count=3,
                thid=testing_thid
            )
            # open
            await ro.open()
            # checks read data md5
            read_data = await ro.read()
            assert calc_bytes_hash(read_data) == file_under_test_md5
            # check EOF
            with pytest.raises(StreamEOF):
                await ro.seek_to_chunk(1000)
            await ro.seek_to_chunk(ro_chunks_num)
            with pytest.raises(StreamEOF):
                await ro.read_chunk()
            # close
            await ro.close()

    results = await run_coroutines(called(), caller(), timeout=15)
    assert len(results) == 2


@pytest.mark.asyncio
async def test_readonly_stream_protocol_persistance(files_dir: str, config_c: dict, config_d: dict):
    """Check readonly stream persistent mechanism
        Agent-C is CALLER
        Agent-D is CALLED
        """
    caller_p2p = await get_pairwise3(me=config_c, their=config_d)
    called_p2p = await get_pairwise3(me=config_d, their=config_c)
    testing_thid = 'test-readonly-protocol-' + uuid.uuid4().hex
    persist_id = 'persist_' + uuid.uuid4().hex
    file_under_test = os.path.join(files_dir, 'big_img.jpeg')
    file_under_test_md5 = calc_file_hash(file_under_test)
    ro_chunks_num = 10

    async def called():
        async with sirius_sdk.context(**config_d):
            co = await sirius_sdk.spawn_coprotocol()
            await co.subscribe(testing_thid)
            try:
                while True:
                    # Pull message
                    e = await co.get_message()
                    # Every time create empty state machine instance
                    state_machine = CalledReadOnlyStreamProtocol(
                        called_p2p,
                        thid=testing_thid,
                        persistent_id=persist_id,
                        # create new stream instance (with empty state)
                        proxy_to=FileSystemReadOnlyStream(path=file_under_test, chunks_num=ro_chunks_num)
                    )
                    # load state
                    await state_machine.load()
                    await state_machine.handle(e.message)
                    if state_machine.edited:
                        # save state if need
                        await state_machine.save()
                    if e.message.operation == StreamOperation.OperationCode.CLOSE:
                        await state_machine.abort()
                        return
            finally:
                await co.abort()

    async def caller():
        async with sirius_sdk.context(**config_c):
            ro = CallerReadOnlyStreamProtocol(
                called=caller_p2p, uri=file_under_test, read_timeout=3, retry_count=3,
                thid=testing_thid
            )
            # open
            await ro.open()
            # checks read data md5
            read_data = await ro.read()
            assert calc_bytes_hash(read_data) == file_under_test_md5
            # close
            await ro.close()

    results = await run_coroutines(called(), caller(), timeout=15)
    assert len(results) == 2


@pytest.mark.asyncio
async def test_writeonly_stream_protocols(config_c: dict, config_d: dict):
    """
    Agent-C is CALLER
    Agent-D is CALLED
    """

    caller_p2p = await get_pairwise3(me=config_c, their=config_d)
    called_p2p = await get_pairwise3(me=config_d, their=config_c)
    testing_thid = 'test-writeonly-protocol-' + uuid.uuid4().hex
    file_under_test = os.path.join(tempfile.tempdir, f'{uuid.uuid4().hex}.bin')
    chunks_num = 5
    chunk_size = 128
    etalon_chunks = []
    for n in range(chunks_num):
        s = f'{n}' * chunk_size
        etalon_chunks.append(s.encode())
    s = 'x' * (chunk_size // 2)
    etalon_chunks.append(s.encode())
    etalon_data = b''
    for b in etalon_chunks:
        etalon_data += b
    etalon_md5 = calc_bytes_hash(etalon_data)
    print('')

    async def called():
        async with sirius_sdk.context(**config_d):
            stream = FileSystemWriteOnlyStream(path=file_under_test, chunk_size=chunk_size)
            await stream.create()
            try:
                state_machine = CalledWriteOnlyStreamProtocol(called_p2p, thid=testing_thid)
                await state_machine.run_forever(stream)
            finally:
                os.remove(file_under_test)

    async def caller():
        async with sirius_sdk.context(**config_c):
            wo = CallerWriteOnlyStreamProtocol(called=caller_p2p, uri=file_under_test, thid=testing_thid)
            # open
            await wo.open()
            # checks...
            assert wo.is_open is True
            assert wo.current_chunk == 0
            assert wo.chunks_num == 0
            assert wo.seekable is True
            assert wo.chunk_size == chunk_size
            # write all
            for no, chunk in enumerate(etalon_chunks):
                new_no, sz = await wo.write_chunk(chunk, no)
                assert new_no == no + 1
                assert sz == len(chunk)
                assert wo.current_chunk == new_no
            # check MD5
            md5 = calc_file_hash(file_under_test)
            assert etalon_md5 == md5
            # close
            await wo.close()
            assert wo.is_open is False

    results = await run_coroutines(called(), caller(), timeout=5)
    assert len(results) == 2


@pytest.mark.asyncio
async def test_writeonly_stream_protocol_persistance(config_c: dict, config_d: dict):
    """Check writeonly stream persistent mechanism
    Agent-C is CALLER
    Agent-D is CALLED
    """

    caller_p2p = await get_pairwise3(me=config_c, their=config_d)
    called_p2p = await get_pairwise3(me=config_d, their=config_c)
    testing_thid = 'test-writeonly-protocol-' + uuid.uuid4().hex
    file_under_test = os.path.join(tempfile.tempdir, f'{uuid.uuid4().hex}.bin')
    etalon_data = b'x' * 1024 * 3
    persist_id = uuid.uuid4().hex
    etalon_md5 = calc_bytes_hash(etalon_data)
    print('')

    async def called():
        async with sirius_sdk.context(**config_d):
            stream = FileSystemWriteOnlyStream(path=file_under_test)
            await stream.create()
            co = await sirius_sdk.spawn_coprotocol()
            await co.subscribe(testing_thid)
            try:
                while True:
                    # Pull message
                    e = await co.get_message()
                    # Every time create empty state machine instance
                    state_machine = CalledWriteOnlyStreamProtocol(
                        called_p2p,
                        thid=testing_thid,
                        persistent_id=persist_id,
                        # create new stream instance (with empty state)
                        proxy_to=FileSystemWriteOnlyStream(path=file_under_test)
                    )
                    # load state
                    await state_machine.load()
                    await state_machine.handle(e.message)
                    if state_machine.edited:
                        # save state if need
                        await state_machine.save()
                    if e.message.operation == StreamOperation.OperationCode.CLOSE:
                        await state_machine.abort()
                        return
            finally:
                await co.abort()
                os.remove(file_under_test)

    async def caller():
        async with sirius_sdk.context(**config_c):
            wo = CallerWriteOnlyStreamProtocol(called=caller_p2p, uri=file_under_test, thid=testing_thid)
            # open
            await wo.open()
            await wo.write(etalon_data)
            # check MD5
            md5 = calc_file_hash(file_under_test)
            assert etalon_md5 == md5
            # close
            await wo.close()
            assert wo.is_open is False

    results = await run_coroutines(called(), caller(), timeout=5)
    assert len(results) == 2


@pytest.mark.asyncio
async def test_writeonly_stream_protocol_write_all(config_c: dict, config_d: dict):
    """
    Agent-C is CALLER
    Agent-D is CALLED
    """

    caller_p2p = await get_pairwise3(me=config_c, their=config_d)
    called_p2p = await get_pairwise3(me=config_d, their=config_c)
    testing_thid = 'test-writeonly-protocol-' + uuid.uuid4().hex
    file_under_test = os.path.join(tempfile.tempdir, f'{uuid.uuid4().hex}.bin')
    chunk_size = 128
    data_size = chunk_size * 5 + 100
    data = b'0' * data_size
    etalon_md5 = calc_bytes_hash(data)
    print('')

    async def called():
        async with sirius_sdk.context(**config_d):
            stream = FileSystemWriteOnlyStream(path=file_under_test, chunk_size=chunk_size)
            await stream.create()
            try:
                state_machine = CalledWriteOnlyStreamProtocol(called_p2p, thid=testing_thid)
                await state_machine.run_forever(stream)
            finally:
                os.remove(file_under_test)

    async def caller():
        async with sirius_sdk.context(**config_c):
            wo = CallerWriteOnlyStreamProtocol(called=caller_p2p, uri=file_under_test, thid=testing_thid)
            # open
            await wo.open()
            # write all
            await wo.write(data)
            # check MD5
            md5 = calc_file_hash(file_under_test)
            assert etalon_md5 == md5
            # close
            await wo.close()

    results = await run_coroutines(called(), caller(), timeout=5)
    assert len(results) == 2


@pytest.mark.asyncio
async def test_stream_protocols_encoding(config_c: dict, config_d: dict):
    """
    Agent-C is CALLER
    Agent-D is CALLED
    """

    caller_p2p = await get_pairwise3(me=config_c, their=config_d)
    called_p2p = await get_pairwise3(me=config_d, their=config_c)
    testing_thid = 'test-writeonly-protocol-encoding-' + uuid.uuid4().hex
    file_under_test = os.path.join(tempfile.tempdir, f'{uuid.uuid4().hex}.bin')

    recip_vk_bytes, recip_sigkey_bytes = create_keypair(b'00000000000000000000000RECIPIENT')
    recip_vk, recip_sigkey = bytes_to_b58(recip_vk_bytes), bytes_to_b58(recip_sigkey_bytes)
    enc = StreamEncryption().setup(target_verkeys=[recip_vk])
    dec = StreamDecryption(recipients=enc.recipients, nonce=enc.nonce)
    dec.setup(recip_vk, recip_sigkey)
    data_size = 5000
    data = b'0' * data_size

    async def called():
        async with sirius_sdk.context(**config_d):
            wo = FileSystemWriteOnlyStream(
                path=file_under_test
            )
            await wo.create()
            try:
                state_machine1 = CalledWriteOnlyStreamProtocol(called_p2p, thid=testing_thid)
                await state_machine1.run_forever(wo)

                state_machine2 = CalledReadOnlyStreamProtocol(called_p2p, thid=testing_thid)
                ro = FileSystemReadOnlyStream(
                    path=file_under_test,
                    chunks_num=wo.chunks_num
                )
                await state_machine2.run_forever(ro)
            finally:
                os.remove(file_under_test)

    async def caller():
        async with sirius_sdk.context(**config_c):
            nonlocal data
            # Write data
            wo = CallerWriteOnlyStreamProtocol(
                called=caller_p2p,
                uri=file_under_test,
                thid=testing_thid,
                enc=enc
            )
            # open
            await wo.open()
            # write all
            await wo.write(data)
            # close
            await wo.close()
            # sleep
            await asyncio.sleep(1)
            # reader
            ro = CallerReadOnlyStreamProtocol(
                called=caller_p2p,
                uri=file_under_test,
                thid=testing_thid,
                enc=dec,
                read_timeout=3
            )
            await ro.open()
            actual_data = await ro.read()
            await ro.close()
            assert actual_data == data

    results = await run_coroutines(called(), caller(), timeout=5)
    assert len(results) == 2


@pytest.mark.asyncio
async def test_stream_protocols_truncate(config_c: dict, config_d: dict):
    """
    Agent-C is CALLER
    Agent-D is CALLED
    """

    caller_p2p = await get_pairwise3(me=config_c, their=config_d)
    called_p2p = await get_pairwise3(me=config_d, their=config_c)
    testing_thid = 'test-writeonly-protocol-encoding-' + uuid.uuid4().hex
    file_under_test = os.path.join(tempfile.tempdir, f'{uuid.uuid4().hex}.bin')

    recip_vk_bytes, recip_sigkey_bytes = create_keypair(b'00000000000000000000000RECIPIENT')
    recip_vk, recip_sigkey = bytes_to_b58(recip_vk_bytes), bytes_to_b58(recip_sigkey_bytes)
    enc = StreamEncryption().setup(target_verkeys=[recip_vk])
    dec = StreamDecryption(recipients=enc.recipients, nonce=enc.nonce)
    dec.setup(recip_vk, recip_sigkey)
    chunks = [b'chunk1', b'chunk2', b'chunk3']
    content = ''.join([s.decode() for s in chunks]).encode()

    async def called():
        async with sirius_sdk.context(**config_d):
            wo = FileSystemWriteOnlyStream(
                path=file_under_test
            )
            await wo.create()
            try:
                state_machine1 = CalledWriteOnlyStreamProtocol(called_p2p, thid=testing_thid)
                await state_machine1.run_forever(wo)

                state_machine2 = CalledReadOnlyStreamProtocol(called_p2p, thid=testing_thid)
                ro = FileSystemReadOnlyStream(
                    path=file_under_test,
                    chunks_num=wo.chunks_num
                )
                await state_machine2.run_forever(ro)
            finally:
                os.remove(file_under_test)

    async def caller():
        async with sirius_sdk.context(**config_c):
            nonlocal chunks
            # Write data
            wo = CallerWriteOnlyStreamProtocol(
                called=caller_p2p,
                uri=file_under_test,
                thid=testing_thid,
                enc=enc
            )
            # Write
            await wo.open()
            for chunk in chunks:
                await wo.write_chunk(chunk)
            await wo.truncate(1)
            await wo.close()
            # sleep
            await asyncio.sleep(1)
            # reader
            ro = CallerReadOnlyStreamProtocol(
                called=caller_p2p,
                uri=file_under_test,
                thid=testing_thid,
                enc=dec,
                read_timeout=3
            )
            await ro.open()
            actual_data = await ro.read()
            await ro.close()
            assert actual_data == chunks[0]

    results = await run_coroutines(called(), caller(), timeout=5)
    assert len(results) == 2


@pytest.mark.asyncio
async def test_encrypted_documents(config_a: dict, config_b: dict, config_c: dict):
    """
    Agent-A + Agent-B are reader parties
    Agent-C is publisher party
    """

    # Recipient-1
    async with sirius_sdk.context(**config_a):
        recip_vk_a = await sirius_sdk.Crypto.create_key()
    # Recipient-1
    async with sirius_sdk.context(**config_b):
        recip_vk_b = await sirius_sdk.Crypto.create_key()
    # Sender
    async with sirius_sdk.context(**config_c):
        sender_vk = await sirius_sdk.Crypto.create_key()

    content = 'some-data-and-non-ascii-aåbäcö'.encode()
    document = EncryptedDocument(target_verkeys=[recip_vk_a, recip_vk_b])

    document.content = content
    assert document.content == content
    assert document.encrypted is False
    # Publisher
    async with sirius_sdk.context(**config_c):
        await document.encrypt(my_vk=sender_vk)

    assert document.content != content
    assert document.encrypted is True

    for party in [config_a, config_b]:
        async with sirius_sdk.context(**party):
            tmp = EncryptedDocument(src=document)
            await tmp.decrypt()
            assert tmp.sender_vk == sender_vk
            assert tmp.content == content
            assert set(tmp.target_verkeys) == set([recip_vk_a, recip_vk_b])
            await tmp.encrypt()  # encrypt again


@pytest.mark.asyncio
async def test_encrypted_documents_save_load(config_a: dict):
    """
    Agent-A is reader parties
    """
    # Recipient
    async with sirius_sdk.context(**config_a):
        recip_vk = await sirius_sdk.Crypto.create_key()

    expected_content = 'some-data-and-non-ascii-aåbäcö'.encode()
    document = EncryptedDocument(target_verkeys=[recip_vk])
    document.content = expected_content

    file_under_test = os.path.join(tempfile.tempdir, f'{uuid.uuid4().hex}.bin')
    wo = FileSystemWriteOnlyStream(file_under_test)
    await wo.create(truncate=True)
    try:
        #
        await wo.open()
        try:
            await document.save(wo)
            chunks_num = wo.chunks_num
        finally:
            await wo.close()
        ro = FileSystemReadOnlyStream(file_under_test, chunks_num)
        await ro.open()
        try:
            loaded_doc = EncryptedDocument()
            await loaded_doc.load(ro)
            assert loaded_doc.content == expected_content
        finally:
            await ro.close()
    finally:
        os.remove(file_under_test)


@pytest.mark.asyncio
async def test_recipe_simple_vault_init(config_a: dict, config_b: dict):
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    p2p = await get_pairwise3(me=config_a, their=config_b)
    os.mkdir(dir_under_test)
    try:
        # Storage Hub side
        auth = ConfidentialStorageAuthProvider()
        # Check auth checks
        with pytest.raises(ConfidentialStoragePermissionDenied):
            SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
        await auth.authorize(p2p)
        assert auth.authorized is True
        # Check init
        vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
        assert vault.cfg is not None
        assert vault.cfg.as_json() != {}
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_recipe_simple_datavault_stream_operations(config_a: dict, config_b: dict):
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    p2p = await get_pairwise3(me=config_a, their=config_b)
    os.mkdir(dir_under_test)
    path_under_test = f'stream_{uuid.uuid4().hex}.bin'
    test_data = b'x' * 1024 * 5
    try:
        # Storage Hub side
        async with sirius_sdk.context(**config_a):
            auth = ConfidentialStorageAuthProvider()
            await auth.authorize(p2p)
            vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
            await vault.open()
            print('#1')
            doc = await vault.create_stream(path_under_test)
            uri_under_test = doc.id
            print('#2')
            stream_to_write = await vault.writable(uri_under_test)
            await stream_to_write.open()
            try:
                await stream_to_write.write(test_data)
            finally:
                await stream_to_write.close()
            print('#3')
            stream_to_read = await vault.readable(uri_under_test)
            assert stream_to_read.chunks_num > 0
            await stream_to_read.open()
            try:
                read_data = await stream_to_read.read()
            finally:
                await stream_to_read.close()
            print('#4')
            assert read_data == test_data
            file_uri = stream_to_read.path
            print('#5')
        # Check data is encrypted
        with open(file_uri, 'rb') as f:
            raw_content = f.read()
        assert raw_content != test_data
        await vault.close()
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_recipe_simple_datavault_docs_operations(config_a: dict, config_b: dict):
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    p2p = await get_pairwise3(me=config_a, their=config_b)
    os.mkdir(dir_under_test)
    path_under_test = f'stream_{uuid.uuid4().hex}.bin'
    test_data = b'Some user content'
    try:
        # Storage Hub side
        async with sirius_sdk.context(**config_a):
            auth = ConfidentialStorageAuthProvider()
            await auth.authorize(p2p)
            vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
            await vault.open()
            print('#1')
            doc = await vault.create_document(path_under_test)
            uri_under_test = doc.id
            print('#2')
            # Store non-encrypted doc
            doc = EncryptedDocument()
            doc.content = test_data
            await vault.save_document(uri_under_test, doc)
            print('#3')
            loaded_doc = await vault.load(uri_under_test)
            assert isinstance(loaded_doc, StructuredDocument)
            assert isinstance(loaded_doc.content, EncryptedDocument)
            assert loaded_doc.content.content == test_data
            assert loaded_doc.content.encrypted is False
            # Encrypted doc
            async with sirius_sdk.context(**config_b):
                recipient_vk = await sirius_sdk.Crypto.create_key()
            enc_doc = EncryptedDocument(target_verkeys=[recipient_vk])
            enc_doc.content = test_data
            print('#4')
            await enc_doc.encrypt()
            await vault.save_document(uri_under_test, enc_doc)
            print('#5')
            loaded_enc_doc = await vault.load(uri_under_test)
            assert isinstance(loaded_enc_doc, StructuredDocument)
            assert isinstance(loaded_enc_doc.content, EncryptedDocument)
            assert loaded_enc_doc.content.encrypted is True
            assert loaded_enc_doc.content.content != test_data
            async with sirius_sdk.context(**config_b):
                await loaded_enc_doc.content.decrypt()
                assert loaded_enc_doc.content.content == test_data
            await vault.close()
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_recipe_simple_datavault_metadata_and_indexes(config_a: dict, config_b: dict):
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    p2p = await get_pairwise3(me=config_a, their=config_b)
    os.mkdir(dir_under_test)
    test_data = b'Some user content'
    path_under_test1 = f'document_{uuid.uuid4().hex}.bin'
    path_under_test2 = f'stream_{uuid.uuid4().hex}.bin'
    attr1_val = 'attr1-' + uuid.uuid4().hex
    attr2_val = 'attr2-' + uuid.uuid4().hex
    try:
        # Storage Hub side
        async with sirius_sdk.context(**config_a):
            auth = ConfidentialStorageAuthProvider()
            await auth.authorize(p2p)
            vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
            await vault.open()
            doc = await vault.create_document(path_under_test1, meta={'meta1': 'value1'}, attr1=attr1_val)
            uri_under_test1 = doc.id
            stream = await vault.create_stream(path_under_test2, meta={'contentType': 'video/mpeg'}, attr2=attr2_val)
            uri_under_test2 = stream.id
            # check1
            load_for_doc = await vault.load(uri_under_test1)
            assert isinstance(load_for_doc, StructuredDocument)
            assert isinstance(load_for_doc.content, EncryptedDocument)
            assert load_for_doc.meta['meta1'] == 'value1'
            assert 'created' in load_for_doc.meta
            assert load_for_doc.urn
            assert load_for_doc.indexed and load_for_doc.indexed[0].attributes == ['attr1']
            # check2
            load_for_stream = await vault.load(uri_under_test2)
            assert isinstance(load_for_stream, StructuredDocument)
            ro = await load_for_stream.stream.readable()
            assert isinstance(ro, AbstractReadOnlyStream)
            wo = await load_for_stream.stream.writable()
            assert isinstance(wo, AbstractWriteOnlyStream)
            assert load_for_stream.meta['contentType'] == 'video/mpeg'
            assert 'created' in load_for_stream.meta
            assert 'chunks' in load_for_stream.meta and type(load_for_stream.meta['chunks']) is int
            assert load_for_stream.urn
            assert load_for_doc.indexed and load_for_stream.indexed[0].attributes == ['attr2']
            stream_to_write = await vault.writable(uri_under_test2)
            await stream_to_write.open()
            try:
                await stream_to_write.write(test_data)
            finally:
                await stream_to_write.close()
            load_for_stream2 = await vault.load(uri_under_test2)
            assert load_for_stream2.meta['chunks'] > 0
            # Check indexes
            indexes = await vault.indexes()
            assert indexes
            collection1 = await indexes.filter(attr1=attr1_val)
            assert len(collection1) == 1
            collection2 = await indexes.filter(attr2=attr2_val)
            assert len(collection2) == 1
            collection3 = await indexes.filter(attr1=attr1_val, attr2=attr2_val)
            assert len(collection3) == 0
            await vault.close()
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_recipe_simple_datavault_update_metadata_and_attributes(config_a: dict, config_b: dict):
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    p2p = await get_pairwise3(me=config_a, their=config_b)
    os.mkdir(dir_under_test)
    path_under_test_doc = f'document_{uuid.uuid4().hex}.bin'
    path_under_test_stream = f'stream_{uuid.uuid4().hex}.bin'
    attr1_val = 'attr1'
    attr2_val = 'attr2'
    meta1 = 'meta1'
    meta2 = 'meta2'
    try:
        # Storage Hub side
        async with sirius_sdk.context(**config_a):
            auth = ConfidentialStorageAuthProvider()
            await auth.authorize(p2p)
            vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
            await vault.open()
            # Create
            doc = await vault.create_document(path_under_test_doc, meta={'meta1': meta1}, attr1=attr1_val)
            uri_under_test_doc = doc.id
            stream = await vault.create_stream(path_under_test_stream, meta={'meta1': meta1}, attr1=attr1_val)
            uri_under_test_stream = stream.id
            # Update
            await vault.update(uri_under_test_doc, meta={'meta2': meta2}, attr2=attr2_val)
            await vault.update(uri_under_test_stream, meta={'meta2': meta2, 'contentType': 'video/mpeg'}, attr2=attr2_val)
            # Checks
            load_for_doc = await vault.load(uri_under_test_doc)
            assert load_for_doc.meta['meta2'] == meta2
            assert 'meta1' not in load_for_doc.meta
            assert 'created' in load_for_doc.meta
            assert load_for_doc.urn
            assert load_for_doc.indexed and load_for_doc.indexed[0].attributes == ['attr2']
            # check2
            load_for_stream = await vault.load(uri_under_test_stream)
            assert load_for_stream.meta['meta2'] == meta2
            assert 'meta1' not in load_for_stream.meta
            assert 'created' in load_for_stream.meta
            assert 'chunks' in load_for_stream.meta
            assert load_for_stream.urn
            assert load_for_stream.meta['contentType'] == 'video/mpeg'
            assert load_for_stream.indexed and load_for_doc.indexed[0].attributes == ['attr2']
            await vault.close()
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_recipe_simple_datavault_access_with_url_and_urn(config_a: dict, config_b: dict):
    """Check user can load documents/streams both with URI and URN value"""
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    p2p = await get_pairwise3(me=config_a, their=config_b)
    os.mkdir(dir_under_test)
    path_under_test_doc = f'document_{uuid.uuid4().hex}.bin'
    path_under_test_stream = f'stream_{uuid.uuid4().hex}.bin'
    try:
        # Storage Hub side
        async with sirius_sdk.context(**config_a):
            auth = ConfidentialStorageAuthProvider()
            await auth.authorize(p2p)
            vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
            await vault.open()
            # Create
            created_doc = await vault.create_document(path_under_test_doc)
            uri_under_test_doc = created_doc.id
            # Load with URI
            doc_via_uri = await vault.load(uri_under_test_doc)
            doc_via_urn = await vault.load(doc_via_uri.urn)
            assert doc_via_uri.id == doc_via_urn.id == created_doc.id
            assert doc_via_uri.urn == doc_via_urn.urn == created_doc.urn
            # Repeat ones for Stream
            created_stream = await vault.create_document(path_under_test_stream)
            uri_under_test_stream = created_stream.id
            stream_via_uri = await vault.load(uri_under_test_stream)
            stream_via_urn = await vault.load(stream_via_uri.urn)
            assert stream_via_uri.id == stream_via_urn.id == created_stream.id
            assert stream_via_uri.urn == stream_via_urn.urn == created_stream.urn
            await vault.close()
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_structured_document_attach(config_a: dict):
    sd = StructuredDocumentAttach(id='document-id')
    async with sirius_sdk.context(**config_a):
        js = {
            "id": "urn:uuid:41289468-c42c-4b28-adb0-bf76044aec77",
            "sequence": 0,
            "indexed": [
                {
                    "sequence": 0,
                    "attributes": [
                    ]
                }
            ],
            "meta": {
                "created": "2019-06-19",
                "contentType": "video/mpeg",
                "chunks": 16,
            },
            "stream": {
                "id": "https://example.com/encrypted-data-vaults/zMbxmSDn2Xzz?hl=zb47JhaKJ3hJ5Jkw8oan35jK23289Hp",
            },
            "content": {
                "message": "Hello World!"
            }
        }
        sd.from_json(js)
        assert sd.id == 'https://example.com/encrypted-data-vaults/zMbxmSDn2Xzz?hl=zb47JhaKJ3hJ5Jkw8oan35jK23289Hp'
        assert sd.urn == 'urn:uuid:41289468-c42c-4b28-adb0-bf76044aec77'
        assert sd.sequence == 0
        assert sd.meta.created == '2019-06-19'
        assert sd.meta.content_type == 'video/mpeg'
        assert sd.meta.chunks == 16
        assert sd.stream.id == 'https://example.com/encrypted-data-vaults/zMbxmSDn2Xzz?hl=zb47JhaKJ3hJ5Jkw8oan35jK23289Hp'
        assert sd.document.encrypted is False
        assert sd.document.content == "Hello World!"

        target_vk = await sirius_sdk.Crypto.create_key()
        packed = await sirius_sdk.Crypto.pack_message({"message": "Hello World!"}, recipient_verkeys=[target_vk])
        js = {
            "id": "urn:uuid:41289468-c42c-4b28-adb0-bf76044aec77",
            "jwm": json.loads(packed.decode())
        }
        sd.from_json(js)
        assert sd.id == 'urn:uuid:41289468-c42c-4b28-adb0-bf76044aec77'
        assert sd.sequence is None
        assert sd.meta is None
        assert sd.stream is None
        assert sd.document.encrypted is True
        await sd.document.decrypt()
        assert sd.document.content == {"message": "Hello World!"}


@pytest.mark.asyncio
async def test_simple_datavault_critical_ops(config_a: dict, config_b: dict):
    p2p = await get_pairwise3(me=config_a, their=config_b)
    path = f'doc_{uuid.uuid4().hex}.bin'
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    os.mkdir(dir_under_test)
    try:
        async with sirius_sdk.context(**config_a):
            auth = ConfidentialStorageAuthProvider()
            await auth.authorize(p2p)
            # Init and configure Vault
            vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
            await vault.open()
            # Case-1: create doc twice
            doc = await vault.create_document(path)
            assert doc.id.startswith('file://')
            uri = doc.id
            # check uri has absolute nature
            with pytest.raises(DataVaultCreateResourceError):
                await vault.create_document(uri)
            await vault.load(uri)
            # Case-2: remove file physically but meta cached in storage
            path_under_test = os.path.join(dir_under_test, vault.mounted_dir, path)
            os.remove(path_under_test)
            with pytest.raises(DataVaultResourceMissing):
                await vault.load(uri)
            await vault.create_document(uri)
            await vault.load(uri)
            # Case-3: remove
            await vault.remove(uri)
            assert os.path.isfile(path_under_test) is False
            with pytest.raises(DataVaultResourceMissing):
                await vault.load(uri)
            # Case-4: Relative path
            for uri in [f'/dir/subdir/doc_{uuid.uuid4().hex}.bin', f'dir/subdir/doc_{uuid.uuid4().hex}.bin']:
                _ = await vault.create_document(uri)
                assert uri in _.id
            # Clear
            await vault.close()
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_simple_datavault_reopen(config_a: dict, config_b: dict):
    p2p = await get_pairwise3(me=config_a, their=config_b)
    path = f'doc_{uuid.uuid4().hex}.bin'
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    os.mkdir(dir_under_test)
    try:
        async with sirius_sdk.context(**config_a):
            auth = ConfidentialStorageAuthProvider()
            await auth.authorize(p2p)
            # Init and configure Vault
            vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
            await vault.open()
            # Case-1: create doc twice
            sd = await vault.create_document(path)
            doc = EncryptedDocument(content=b'Test content')
            await vault.save_document(sd.id, doc)
            loaded = await vault.load(sd.id)
            assert loaded.doc.content == b'Test content'
            # Clear
            await vault.close()
            # Reopen
            await vault.open()
            loaded = await vault.load(sd.id)
            assert loaded.doc.content == b'Test content'
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_simple_datavault_multiple_contexts(config_a: dict, config_b: dict):
    p2pa = await get_pairwise3(me=config_a, their=config_b)
    p2pb = await get_pairwise3(me=config_b, their=config_a)
    path = f'doc_{uuid.uuid4().hex}.bin'
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    os.mkdir(dir_under_test)
    try:
        async with sirius_sdk.context(**config_a):
            auth = ConfidentialStorageAuthProvider()
            await auth.authorize(p2pa)
            # Init and configure Vault
            vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
            await vault.open()
            # create doc
            doc1 = await vault.create_document(path)
        async with sirius_sdk.context(**config_b):
            auth = ConfidentialStorageAuthProvider()
            await auth.authorize(p2pb)
            # Init and configure Vault
            vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
            await vault.open()
            # create doc
            doc2 = await vault.create_document(path)

        assert doc1.id == doc2.id

    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_recipe_simple_datavault_encryption_layers_for_documents(config_a: dict, config_b: dict):
    """Check encryption levels:
        - vault encrypt + controller don't encrypt on self side (trust to vault)
        - vault encrypt + controller encrypt data over vault config
        - vault don't encrypt + controller encrypt on self side (it can share encrypted data among pairwises)
        - vault don't encrypt + controller don't encrypt on self side

    P.S. Controller is participant who control data semantic
    """
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    p2p = await get_pairwise3(me=config_a, their=config_b)
    vault_config = config_a
    controller_config = config_b
    test_data = b'xxx' * 1024

    async with sirius_sdk.context(**controller_config):
        controller_encrypt_key = await sirius_sdk.Crypto.create_key()

    try:
        os.mkdir(dir_under_test)
        for case_vault_encrypt, case_controller_encrypt in [(False, False), (False, True), (True, False), (True, True)]:
            path_under_test = f'document_{uuid.uuid4().hex}.bin'
            async with sirius_sdk.context(**vault_config):
                auth = ConfidentialStorageAuthProvider()
                await auth.authorize(p2p)
                # Init and configure Vault
                vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
                if case_vault_encrypt is False:
                    vault.cfg.key_agreement = None
                await vault.open()
                # Test docs
                structured_doc = await vault.create_document(path_under_test)
                structured_doc_file_path = os.path.join(dir_under_test, vault.mounted_dir, path_under_test)
                assert os.path.isfile(structured_doc_file_path)
                if case_controller_encrypt:
                    doc = EncryptedDocument(target_verkeys=[controller_encrypt_key])
                    doc.content = test_data
                    async with sirius_sdk.context(**controller_config):
                        await doc.encrypt()
                else:
                    doc = EncryptedDocument()
                    doc.content = test_data

                await vault.save_document(structured_doc.urn, doc)

                # Case 1
                if case_vault_encrypt is False and case_controller_encrypt is False:
                    # If both parties don't encrypt data so it is raw data stored on disk
                    with open(structured_doc_file_path, 'rb') as f:
                        actual_data = f.read()
                        assert actual_data == test_data
                else:
                    # If vault is encrypted OR controller does encrypt, then data is encrypted on Disk
                    with open(structured_doc_file_path, 'rb') as f:
                        actual_data = f.read()
                        # Data on disk is encrypted
                        assert actual_data != test_data

                # Case 2
                if case_vault_encrypt is True and case_controller_encrypt is False:
                    # Data if read in Vault context is decrypted
                    loaded_doc = await vault.load(structured_doc.id)
                    assert loaded_doc.content.content == test_data

                # Case 3
                if case_vault_encrypt is False and case_controller_encrypt is True:
                    # Doc controller has access to document content
                    loaded_doc = await vault.load(structured_doc.id)
                    async with sirius_sdk.context(**controller_config):
                        await loaded_doc.content.decrypt()
                        assert loaded_doc.content.content == test_data
                        assert loaded_doc.content.encrypted is False
                # Case 4
                if case_vault_encrypt is True and case_controller_encrypt is True:
                    # Vault side don/t have access to data
                    loaded_doc = await vault.load(structured_doc.id)
                    loaded_doc.content.encrypted = True
                    assert loaded_doc.content.content != test_data
                    with pytest.raises(Exception):
                        await loaded_doc.content.decrypt()
                    # Doc controller has access to document content
                    async with sirius_sdk.context(**controller_config):
                        await loaded_doc.content.decrypt()
                        assert loaded_doc.content.content == test_data
                        assert loaded_doc.content.encrypted is False

                await vault.close()
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_recipe_simple_datavault_encryption_layers_for_streams(config_a: dict, config_b: dict):
    """Check encryption levels:
        - vault encrypt + controller don't encrypt on self side (trust to vault)
        - vault encrypt + controller encrypt data over vault config
        - vault don't encrypt + controller encrypt on self side (it can share encrypted data among pairwises)
        - vault don't encrypt + controller don't encrypt on self side

    P.S. Controller is participant who control data semantic
    """
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    p2p = await get_pairwise3(me=config_a, their=config_b)
    vault_config = config_a
    test_data = b'xxx' * 1024
    controller_pk, controller_sk = create_keypair(seed='0000000000000000000000CONTROLLER'.encode())
    controller_pk, controller_sk = bytes_to_b58(controller_pk), bytes_to_b58(controller_sk)

    try:
        os.mkdir(dir_under_test)
        for case_vault_encrypt, case_controller_encrypt in [(False, False), (False, True), (True, False), (True, True)]:
            path_under_test = f'document_{uuid.uuid4().hex}.bin'
            async with sirius_sdk.context(**vault_config):
                auth = ConfidentialStorageAuthProvider()
                await auth.authorize(p2p)
                # Init and configure Vault
                vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
                if case_vault_encrypt is False:
                    vault.cfg.key_agreement = None
                await vault.open()

                # Test stream
                structured_doc = await vault.create_stream(path_under_test)
                structured_doc_file_path = os.path.join(dir_under_test, vault.mounted_dir, path_under_test)
                assert os.path.isfile(structured_doc_file_path)
                stream_to_write_layer1 = await vault.writable(structured_doc.id)
                if case_controller_encrypt:
                    stream_to_write_layer2 = WriteOnlyStreamEncodingWrapper(
                        dest=stream_to_write_layer1,
                        enc=StreamEncryption().setup(target_verkeys=[controller_pk])
                    )
                    jwe_layer2 = stream_to_write_layer2.enc.jwe
                else:
                    # Ignore adv encoding layer
                    stream_to_write_layer2 = stream_to_write_layer1
                    jwe_layer2 = None
                # Write Data to stream under test
                await stream_to_write_layer2.open()
                await stream_to_write_layer2.write(test_data)
                await stream_to_write_layer2.close()

                # Run Test cases

                # Case 1
                if case_vault_encrypt is False and case_controller_encrypt is False:
                    # If both parties don't encrypt data so it is raw data stored on disk
                    with open(structured_doc_file_path, 'rb') as f:
                        actual_data = f.read()
                        assert actual_data == test_data
                else:
                    # If vault is encrypted or controller is encrypt, then data is encrypted
                    with open(structured_doc_file_path, 'rb') as f:
                        actual_data = f.read()
                        # Data on disk is encrypted
                        assert actual_data != test_data
                # Case 2
                if case_vault_encrypt is True and case_controller_encrypt is False:
                    # Data if read in Vault context is decrypted
                    loaded_doc = await vault.load(structured_doc.id)
                    stream_to_read_layer2 = await loaded_doc.stream.readable()
                    await stream_to_read_layer2.open()
                    try:
                        actual_data = await stream_to_read_layer2.read()
                    finally:
                        await stream_to_read_layer2.close()
                    assert actual_data == test_data
                # Case 3
                if case_vault_encrypt is False and case_controller_encrypt is True:
                    # Doc controller has access to document content
                    loaded_doc = await vault.load(structured_doc.id)
                    stream_to_read_layer2 = await loaded_doc.stream.readable(
                        jwe_layer2,
                        keys=KeyPair(controller_pk, controller_sk)
                    )
                    await stream_to_read_layer2.open()
                    try:
                        actual_data = await stream_to_read_layer2.read()
                    finally:
                        await stream_to_read_layer2.close()
                    assert actual_data == test_data
                # Case 4
                if case_vault_encrypt is True and case_controller_encrypt is True:
                    # Vault side don/t have access to data
                    loaded_doc = await vault.load(structured_doc.id)
                    stream_to_read_layer2 = await loaded_doc.stream.readable(
                        jwe_layer2,
                        keys=KeyPair(controller_pk, controller_sk)
                    )
                    await stream_to_read_layer2.open()
                    try:
                        actual_data = await stream_to_read_layer2.read()
                    finally:
                        await stream_to_read_layer2.close()
                    assert actual_data == test_data

                await vault.close()
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_data_vault_state_machines_for_docs(config_a: dict, config_b: dict):
    vault_cfg = config_a
    controller_cfg = config_b
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    caller = await get_pairwise3(me=controller_cfg, their=vault_cfg)
    called = await get_pairwise3(me=vault_cfg, their=controller_cfg)
    os.mkdir(dir_under_test)
    try:
        auth = ConfidentialStorageAuthProvider()
        await auth.authorize(called)
        # Init and configure Vault
        vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)

        async def run_vault():
            sm = CalledEncryptedDataVault(called, proxy_to=[vault])
            async with sirius_sdk.context(**vault_cfg):
                async for e in (await sirius_sdk.subscribe()):
                    assert e.pairwise.their.did == called.their.did
                    assert isinstance(e.message, BaseConfidentialStorageMessage)
                    request: BaseConfidentialStorageMessage = e.message
                    print('')
                    await sm.handle(request)
                    print('')

        async def run_controller():
            async with sirius_sdk.context(**controller_cfg):
                m = CallerEncryptedDataVault(caller)
                print('List all vaults')
                vaults = await m.list_vaults()
                assert vaults
                vault_id = vaults[0].id
                print('Select Vault')
                m.select(vault_id)
                print('Open selected vault')
                await m.open()
                print('Create resources')
                sd1 = await m.create_document(uri=f'my_document.bin', meta={'meta1': 'value-1'}, attr1='attr1')
                assert 'my_document.bin' in sd1.id
                assert sd1.doc is not None and sd1.doc.content == b'' and sd1.doc.encrypted is False
                assert sd1.urn.startswith('urn:uuid:')
                assert sd1.meta['meta1'] == 'value-1'
                assert sd1.indexed[0].attributes == ['attr1']
                print('Update resource metadata')
                await m.update(sd1.id, meta={'meta-x': 'value-x'}, attrx='attrx')
                loaded = await m.load(sd1.id)
                assert loaded.meta['meta-x'] == 'value-x'
                assert loaded.indexed[0].attributes == ['attrx']
                assert loaded.id == sd1.id
                assert loaded.urn == sd1.urn
                print('Chek document save/load/encrypt/decrypt operations')
                # save/load non-encrypted doc
                doc = EncryptedDocument()
                doc.content = 'Test message'
                await m.save_document(sd1.id, doc)
                loaded2 = await m.load(sd1.id)
                assert loaded2.doc.content == 'Test message'
                assert loaded2.doc.encrypted is False
                # save/load encrypted doc
                target_vk = await sirius_sdk.Crypto.create_key()
                enc = EncryptedDocument(target_verkeys=[target_vk])
                enc.content = b'Test message'
                await enc.encrypt()
                print('')
                await m.save_document(sd1.id, enc)
                loaded3 = await m.load(sd1.id)
                assert loaded3.doc.encrypted is True
                await loaded3.doc.decrypt()
                assert loaded3.doc.content == b'Test message'
                # Finish
                await m.close()

        results = await run_coroutines(
            run_vault(),
            run_controller(),
            timeout=5
        )
        assert results
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_data_vault_state_machines_for_streams(config_a: dict, config_b: dict):
    vault_cfg = config_a
    controller_cfg = config_b
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    caller = await get_pairwise3(me=controller_cfg, their=vault_cfg)
    called = await get_pairwise3(me=vault_cfg, their=controller_cfg)
    os.mkdir(dir_under_test)
    try:
        auth = ConfidentialStorageAuthProvider()
        await auth.authorize(called)
        # Init and configure Vault
        vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)

        async def run_vault():
            sm = CalledEncryptedDataVault(called, proxy_to=[vault])
            async with sirius_sdk.context(**vault_cfg):
                async for e in (await sirius_sdk.subscribe()):
                    assert e.pairwise.their.did == called.their.did
                    assert isinstance(e.message, BaseConfidentialStorageMessage)
                    request: BaseConfidentialStorageMessage = e.message
                    print('')
                    await sm.handle(request)
                    print('')

        async def run_controller():
            async with sirius_sdk.context(**controller_cfg):
                m = CallerEncryptedDataVault(caller)
                print('List all vaults')
                vaults = await m.list_vaults()
                assert vaults
                vault_id = vaults[0].id
                print('Select Vault')
                m.select(vault_id)
                print('Open selected vault')
                await m.open()
                print('Create resources')
                sd2 = await m.create_stream(uri=f'my_stream.bin', meta={'meta2': 'value-2'}, attr2='attr2')
                assert 'my_stream.bin' in sd2.id
                assert sd2.urn.startswith('urn:uuid:')
                assert sd2.meta['meta2'] == 'value-2'
                assert sd2.indexed[0].attributes == ['attr2']
                print('Update resource metadata')
                await m.update(sd2.id, meta={'meta-x': 'value-x'}, attrx='attrx')
                loaded = await m.load(sd2.id)
                assert loaded.meta['meta-x'] == 'value-x'
                assert loaded.indexed[0].attributes == ['attrx']
                assert loaded.id == sd2.id
                assert loaded.urn == sd2.urn
                # Open stream and write data
                test_data = b'test-data'
                writable = await m.writable(sd2.id)
                await writable.open()
                await writable.write(test_data)
                await writable.close()
                # Open for reading
                readable = await m.readable(sd2.id)
                await readable.open()
                actual_data = await readable.read()
                await readable.close()
                # Check writen is equal to readen bytes
                assert actual_data == test_data
                # Try again for encrypted
                sd_enc = await m.create_stream(uri=f'my_stream_encoded.bin', meta={'meta2': 'value-2'}, attr2='attr2')
                assert sd_enc.stream is not None
                sd_enc = await m.load(sd_enc.id)
                assert sd_enc.stream is not None
                target_vk = await sirius_sdk.Crypto.create_key()
                enc = StreamEncryption().setup(target_verkeys=[target_vk])
                writable_enc = await sd_enc.stream.writable(enc.jwe)
                # !!!! Wrapper encoder JWE will be refreshed !!!!
                wrapper_jwe = writable_enc.enc.jwe
                # !!!!!!!!!!
                await writable_enc.open()
                await writable_enc.write(test_data)
                await writable_enc.close()
                readable_enc = await sd_enc.stream.readable(wrapper_jwe)
                await readable_enc.open()
                actual_data = await readable_enc.read()
                await readable_enc.close()
                assert actual_data == test_data
                # Finish
                await m.close()

        results = await run_coroutines(
            run_vault(),
            run_controller(),
            timeout=5
        )
        assert results
        assert vault.is_open is False
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_data_vault_state_machines_remove_and_indexes(config_a: dict, config_b: dict):
    vault_cfg = config_a
    controller_cfg = config_b
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    caller = await get_pairwise3(me=controller_cfg, their=vault_cfg)
    called = await get_pairwise3(me=vault_cfg, their=controller_cfg)
    os.mkdir(dir_under_test)
    try:
        auth = ConfidentialStorageAuthProvider()
        await auth.authorize(called)
        # Init and configure Vault
        vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)

        async def run_vault():
            sm = CalledEncryptedDataVault(called, proxy_to=[vault])
            async with sirius_sdk.context(**vault_cfg):
                async for e in (await sirius_sdk.subscribe()):
                    assert e.pairwise.their.did == called.their.did
                    assert isinstance(e.message, BaseConfidentialStorageMessage)
                    request: BaseConfidentialStorageMessage = e.message
                    print('')
                    await sm.handle(request)
                    print('')

        async def run_controller():
            async with sirius_sdk.context(**controller_cfg):
                m = CallerEncryptedDataVault(caller)
                print('List all vaults')
                vaults = await m.list_vaults()
                assert vaults
                vault_id = vaults[0].id
                print('Select Vault')
                m.select(vault_id)
                print('Open selected vault')
                await m.open()
                print('Create resources')
                sd1 = await m.create_document('my_document.bin', meta={}, attr1='attr1')
                sd2 = await m.create_stream('my_stream.bin', meta={}, attr1='attr1', attr2='attr2')
                print('Index')
                index = await m.indexes()
                docs1 = await index.filter(attr1='attr1')
                assert len(docs1) == 2
                docs2 = await index.filter(attr1='attr1', attr2='attr2')
                assert len(docs2) == 1
                print('Remove one of resources')
                await m.remove(sd1.id)
                docs3 = await index.filter(attr1='attr1')
                assert len(docs3) == 1
                # Remove by uid
                await m.remove(sd2.urn)
                docs4 = await index.filter(attr1='attr1')
                assert len(docs4) == 0
                # Finish
                await m.close()

        results = await run_coroutines(
            run_vault(),
            run_controller(),
            timeout=5
        )
        assert results
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_data_vault_state_machines_raise_errors(config_a: dict, config_b: dict):
    vault_cfg = config_a
    controller_cfg = config_b
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    caller = await get_pairwise3(me=controller_cfg, their=vault_cfg)
    called = await get_pairwise3(me=vault_cfg, their=controller_cfg)
    os.mkdir(dir_under_test)
    try:
        auth = ConfidentialStorageAuthProvider()
        await auth.authorize(called)
        # Init and configure Vault
        vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)

        async def run_vault():
            sm = CalledEncryptedDataVault(called, proxy_to=[vault])
            async with sirius_sdk.context(**vault_cfg):
                async for e in (await sirius_sdk.subscribe()):
                    assert e.pairwise.their.did == called.their.did
                    assert isinstance(e.message, BaseConfidentialStorageMessage)
                    request: BaseConfidentialStorageMessage = e.message
                    print('')
                    await sm.handle(request)
                    print('')

        async def run_controller():
            async with sirius_sdk.context(**controller_cfg):
                m = CallerEncryptedDataVault(caller)
                print('List all vaults')
                vaults = await m.list_vaults()
                assert vaults
                vault_id = vaults[0].id
                print('Select Vault')
                m.select(vault_id)
                print('Open selected vault')
                await m.open()
                print('#1 Create resource twice')
                await m.create_document('my_document.bin')
                with pytest.raises(DataVaultCreateResourceError):
                    await m.create_document('my_document.bin')
                print('#2 operate with missing resource')
                with pytest.raises(DataVaultResourceMissing):
                    await m.update('file:///missing_document.bin', meta={'meta': 'val'})
                # Finish
                await m.close()

        results = await run_coroutines(
            run_vault(),
            run_controller(),
            timeout=5
        )
        assert results
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_data_vault_recipe_scheduler(config_a: dict, config_b: dict):
    vault_cfg = config_a
    controller_cfg = config_b
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    caller = await get_pairwise3(me=controller_cfg, their=vault_cfg)
    called = await get_pairwise3(me=vault_cfg, their=controller_cfg)
    test_data = b'Test Data'
    os.mkdir(dir_under_test)
    try:
        auth = ConfidentialStorageAuthProvider()
        await auth.authorize(called)
        # Init and configure Vault
        vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)

        async def run_vault():
            async with sirius_sdk.context(**vault_cfg):
                await sirius_sdk.recipes.schedule_vaults(p2p=called, vaults=[vault])

        async def run_controller():
            async with sirius_sdk.context(**controller_cfg):
                m = CallerEncryptedDataVault(caller)
                print('List all vaults')
                vaults = await m.list_vaults()
                assert vaults
                vault_id = vaults[0].id
                print('Select Vault')
                m.select(vault_id)
                print('Open selected vault')
                await m.open()
                # operate with stream
                sd = await m.create_stream('my_stream.bin', meta=StreamMeta(content_type="video/mpeg"))
                wo = await sd.stream.writable()
                await wo.open()
                try:
                    await wo.write(test_data)
                finally:
                    await wo.close()
                ro = await sd.stream.readable()
                await ro.open()
                try:
                    actual_data = await ro.read()
                finally:
                    await ro.close()
                assert actual_data == test_data
                # Finish
                await m.close()

        results = await run_coroutines(
            run_vault(),
            run_controller(),
            timeout=5
        )
        assert results
    finally:
        shutil.rmtree(dir_under_test)


@pytest.mark.asyncio
async def test_data_vault_multilevel_encoding_streams(config_a: dict, config_b: dict):
    vault_cfg = config_a
    controller_cfg = config_b
    dir_under_test = os.path.join(tempfile.tempdir, f'test_vaults_{uuid.uuid4().hex}')
    caller = await get_pairwise3(me=controller_cfg, their=vault_cfg)
    called = await get_pairwise3(me=vault_cfg, their=controller_cfg)
    test_data = b'Test Data'
    os.mkdir(dir_under_test)
    try:
        auth = ConfidentialStorageAuthProvider()
        await auth.authorize(called)
        # Init and configure Vault
        vault = SimpleDataVault(mounted_dir=dir_under_test, auth=auth)
        assert vault.cfg.key_agreement is not None

        async def run_vault():
            async with sirius_sdk.context(**vault_cfg):
                await sirius_sdk.recipes.schedule_vaults(p2p=called, vaults=[vault])

        async def run_controller():
            async with sirius_sdk.context(**controller_cfg):
                m = CallerEncryptedDataVault(caller)
                print('List all vaults')
                vaults = await m.list_vaults()
                assert vaults
                vault_id = vaults[0].id
                print('Select Vault')
                m.select(vault_id)
                print('Open selected vault')
                await m.open()
                # operate with stream
                sd = await m.create_stream('my_stream.bin', meta=StreamMeta(content_type="video/mpeg"))
                # Init encoding
                enc_key = await sirius_sdk.Crypto.create_key()
                encoding = StreamEncryption().setup(target_verkeys=[enc_key])
                wo = await sd.stream.writable(encoding.jwe, encoding.cek)
                await wo.open()
                try:
                    await wo.write(test_data)
                finally:
                    await wo.close()
                ro = await sd.stream.readable(encoding.jwe)
                await ro.open()
                try:
                    actual_data = await ro.read()
                finally:
                    await ro.close()
                assert actual_data == test_data
                # Finish
                await m.close()

        results = await run_coroutines(
            run_vault(),
            run_controller(),
            timeout=5
        )
        assert results
    finally:
        shutil.rmtree(dir_under_test)



import time
import json
import struct
import base64
import logging
from typing import Any
from datetime import datetime, timedelta
from typing import Optional

from pytime import pytime

from sirius_sdk.abstract.api import APICrypto


def utc_to_str(dt: datetime):
    return dt.strftime('%Y-%m-%dT%H:%M:%S') + '+0000'


def str_to_utc(s: str, raise_exceptions: bool = True) -> Optional[datetime]:
    tz_shift = None
    try:
        if '+' in s:
            s, shift = s.split('+')
            tz_shift = timedelta(hours=int(shift))
        s = s.replace('T', ' ')
        ret = pytime.parse(s)
        return ret + tz_shift
    except:
        logging.exception('Error while parse datetime')
        if raise_exceptions:
            raise
        else:
            return None


async def sign(crypto: APICrypto, value: Any, verkey: str, exclude_sig_data: bool = False) -> dict:
    timestamp_bytes = struct.pack(">Q", int(time.time()))

    sig_data_bytes = timestamp_bytes + json.dumps(value).encode('ascii')
    sig_data = base64.urlsafe_b64encode(sig_data_bytes).decode('ascii')

    signature_bytes = await crypto.crypto_sign(verkey, sig_data_bytes)
    signature = base64.urlsafe_b64encode(
        signature_bytes
    ).decode('ascii')

    data = {
        "@type": "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/signature/1.0/ed25519Sha512_single",
        "signer": verkey,
        "signature": signature
    }
    if not exclude_sig_data:
        data['sig_data'] = sig_data

    return data


async def verify_signed(crypto: APICrypto, signed: dict) -> (Any, bool):
    signature_bytes = base64.urlsafe_b64decode(signed['signature'].encode('ascii'))
    sig_data_bytes = base64.urlsafe_b64decode(signed['sig_data'].encode('ascii'))
    sig_verified = await crypto.crypto_verify(
        signed['signer'],
        sig_data_bytes,
        signature_bytes
    )
    data_bytes = base64.urlsafe_b64decode(signed['sig_data'])
    timestamp = struct.unpack(">Q", data_bytes[:8])
    field_json = data_bytes[8:]
    if isinstance(field_json, bytes):
        field_json = field_json.decode('utf-8')
    return json.loads(field_json), sig_verified



import io
import json
import struct
from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import List, Optional, Dict, Union

import nacl.utils
import nacl.bindings

import sirius_sdk
from sirius_sdk.encryption import b58_to_bytes, bytes_to_b58, bytes_to_b64
from sirius_sdk.encryption.ed25519 import prepare_pack_recipient_keys, locate_pack_recipient_key

from .encoding import ConfidentialStorageEncType, JWE, EncRecipient, EncHeader
from .errors import EncryptionError, StreamInitializationError


class BaseStreamEncryption:

    def __init__(self, nonce: str = None, type_: ConfidentialStorageEncType = ConfidentialStorageEncType.X25519KeyAgreementKey2019):
        """"Encryption settings for Streams
        :param nonce: (base58 string) nonce bytes
        """
        if nonce is None:
            nonce = bytes_to_b58(nacl.utils.random(nacl.bindings.crypto_aead_chacha20poly1305_ietf_NPUBBYTES))
        self.__nonce = nonce
        self._recipients = None
        self.__type = type_
        self._cek = None

    @property
    def nonce(self) -> str:
        return self.__nonce

    @nonce.setter
    def nonce(self, value: str):
        self.__nonce = value

    @property
    def recipients(self) -> Optional[Dict]:
        """JWE.recipients document (example: https://identity.foundation/confidential-storage/#example-4-example-encrypted-document)
        """
        return self._recipients

    @property
    def type(self) -> ConfidentialStorageEncType:
        return self.__type

    @property
    def cek(self) -> Optional[bytes]:
        return self._cek

    @property
    def jwe(self) -> Optional[JWE]:
        if self.recipients:
            jwe = JWE(
                iv=self.__nonce,
                recipients=[
                    EncRecipient(
                        encrypted_key=recip['encrypted_key'],
                        header=EncHeader(
                            kid=recip['header']['kid'],
                            sender=recip['header'].get('sender', None),
                            iv=recip['header'].get('iv', None)
                        )
                    )
                    for recip in self.recipients
                ]
            )
            return jwe
        else:
            return None

    @jwe.setter
    def jwe(self, value: Union[JWE, dict]):
        if isinstance(value, JWE):
            js = value.as_json()
            self._recipients = js['recipients']
            self.__nonce = js.get('iv', None)
        elif isinstance(value, dict):
            self._recipients = value['recipients']
            self.__nonce = value.get('iv', None)
        else:
            raise RuntimeError('Unexpected value type')


class StreamEncryption(BaseStreamEncryption):

    def setup(self, target_verkeys: List[str]) -> "StreamEncryption":
        """Prepare for Encryption

        :param target_verkeys: list of base58 encoded target verkeys
        """
        if self.type != ConfidentialStorageEncType.X25519KeyAgreementKey2019:
            raise EncryptionError(f'Unsupported key agreement "{self.type}"')
        jwe_json, cek = prepare_pack_recipient_keys(
            to_verkeys=[b58_to_bytes(key) for key in target_verkeys]
        )
        jwe = json.loads(jwe_json)
        self._recipients = jwe['recipients']
        self._cek = cek
        return self

    @classmethod
    def from_jwe(cls, jwe: Union[JWE, dict], cek: bytes = None) -> "StreamEncryption":
        inst = StreamEncryption()
        inst.jwe = jwe
        inst._cek = cek
        if inst._cek is None and inst.recipients is not None:
            target_verkeys = [item.get('header', {}).get('kid', None) for item in inst.recipients]
            target_verkeys = [key for key in target_verkeys if key is not None]
            if target_verkeys:
                inst.setup(target_verkeys)
        return inst


class StreamDecryption(BaseStreamEncryption):

    def __init__(
            self, recipients: Dict = None, nonce: str = None,
            type_: ConfidentialStorageEncType = ConfidentialStorageEncType.X25519KeyAgreementKey2019
    ):
        super().__init__(nonce, type_)
        self._recipients = recipients

    def setup(self, vk: str, sk: str) -> "StreamDecryption":
        """Prepare for Decryption

        :param vk: (base58 string) decryption verkey
        :param sk: (base58 string) decryption sigkey
        """
        if self.recipients is None:
            raise EncryptionError('Recipients metadata in JWE format expected')
        cek, sender_vk, recip_vk_b58 = locate_pack_recipient_key(
            recipients=self.recipients, my_verkey=b58_to_bytes(vk), my_sigkey=b58_to_bytes(sk)
        )
        self._cek = cek
        return self

    @classmethod
    def from_jwe(cls, jwe: Union[JWE, dict]) -> "StreamDecryption":
        inst = StreamDecryption()
        inst.jwe = jwe
        return inst


class AbstractStream(ABC):

    def __init__(self, path: str, enc: Optional[BaseStreamEncryption] = None):
        """Interface for Low-level layers of Vault Storage

        :param path: path to resource
        :param enc: (optional) encoding config
        """
        self.__path = path
        self.__enc = enc
        self._current_chunk = 0
        self._seekable = None
        self._chunks_num: int = 0
        self._is_open: bool = False

    @property
    def is_open(self) -> bool:
        return self._is_open

    @property
    def path(self) -> str:
        return self.__path

    @property
    def enc(self) -> Optional[BaseStreamEncryption]:
        return self.__enc

    @enc.setter
    def enc(self, value: BaseStreamEncryption):
        self.__enc = value

    @property
    def seekable(self) -> Optional[bool]:
        return self._seekable

    @property
    def chunks_num(self) -> int:
        return self._chunks_num

    @property
    def current_chunk(self) -> int:
        """Stream offset"""
        return self._current_chunk

    @abstractmethod
    async def open(self):
        raise NotImplemented

    @abstractmethod
    async def close(self):
        raise NotImplemented

    @abstractmethod
    async def seek_to_chunk(self, no: int) -> int:
        raise NotImplemented

    async def encrypt(self, chunk: bytes) -> bytes:
        if self.enc:
            if self.enc.type == ConfidentialStorageEncType.UNKNOWN:
                return chunk
            elif self.enc.type == ConfidentialStorageEncType.X25519KeyAgreementKey2019:
                if self.enc.cek:
                    mlen = len(chunk)
                    prefix = struct.pack("i", mlen)
                    nonce_bytes = b58_to_bytes(self.enc.nonce)
                    aad = self._build_aad(self.enc.recipients)
                    encoded = nacl.bindings.crypto_aead_chacha20poly1305_ietf_encrypt(
                        chunk, aad=aad, nonce=nonce_bytes, key=self.enc.cek
                    )
                    payload = prefix + encoded
                    return payload
            else:
                raise EncryptionError('Unknown Encryption Type')
        else:
            return chunk

    async def decrypt(self, chunk: bytes) -> bytes:
        if self.enc:
            if self.enc.type == ConfidentialStorageEncType.UNKNOWN:
                return chunk
            elif self.enc.type == ConfidentialStorageEncType.X25519KeyAgreementKey2019:
                nonce_bytes = b58_to_bytes(self.enc.nonce)
                prefix = chunk[:4]
                payload = chunk[4:]
                mlen = struct.unpack("i", prefix)[0]
                if self.enc.cek:
                    aad = self._build_aad(self.enc.recipients)
                    # Use manually configured encryption settings: [public-key, secret-key]
                    decrypted = nacl.bindings.crypto_aead_chacha20poly1305_ietf_decrypt(
                        ciphertext=payload, aad=aad, nonce=nonce_bytes, key=self.enc.cek
                    )
                else:
                    # Use Wallet to decrypt with previously configured SDK
                    recips = self._build_recips(self.enc.recipients)
                    recips_b64 = bytes_to_b64(json.dumps(recips).encode("ascii"), urlsafe=True)
                    ciphertext = payload[:mlen]
                    tag = payload[mlen:]
                    packed_message = OrderedDict(
                        [
                            ("protected", recips_b64),
                            ("iv", bytes_to_b64(nonce_bytes, urlsafe=True)),
                            ("ciphertext", bytes_to_b64(ciphertext, urlsafe=True)),
                            ("tag", bytes_to_b64(tag, urlsafe=True)),
                        ]
                    )
                    jwe = json.dumps(packed_message).encode()
                    unpacked = await sirius_sdk.Crypto.unpack_message(jwe)
                    decrypted_msg = unpacked['message']
                    if isinstance(decrypted_msg, str):
                        decrypted = decrypted_msg.encode()
                    elif isinstance(decrypted_msg, dict):
                        decrypted = json.dumps(decrypted_msg).encode()
                    else:
                        decrypted = decrypted_msg
                return decrypted
            else:
                raise EncryptionError('Unknown Encryption Type')
        else:
            return chunk

    def _build_recips(self, recipients: dict) -> dict:
        if self.enc.type == ConfidentialStorageEncType.X25519KeyAgreementKey2019:
            recips = OrderedDict(
                [
                    ("enc", "xchacha20poly1305_ietf"),
                    ("typ", "JWM/1.0"),
                    ("alg", "Anoncrypt"),
                    ("recipients", recipients),
                ]
            )
            return recips
        else:
            raise EncryptionError('Unknown encryption format')

    def _build_aad(self, recipients: dict) -> bytes:
        recips = self._build_recips(recipients)
        recips_json = json.dumps(recips)
        recips_b64 = bytes_to_b64(recips_json.encode("ascii"), urlsafe=True)
        aad = recips_b64.encode("ascii")
        return aad


class AbstractReadOnlyStream(AbstractStream):
    """Stream abstraction for reading operations:
      - cloud storage
      - file-system
      - external storages
      - etc
    """

    def __init__(self, path: str, chunks_num: int, enc: Optional[BaseStreamEncryption] = None):
        """
        :param path: path to stream on local infrastructure (device, cloud provider ...)
        :param chunks_num: count of chunks that stream was splitted to
          !!! stream may ignore chunks_num (when stream is encoded for example) !!!
          Chunks allow:
            - partially upload/download big data files/streams (control progress)
            - encrypt/decrypt big data partially
            - adv. services like upload/download with pause/resume (for cloud providers for example)
        :param enc: allow decrypt stream chunks
        """
        super().__init__(path, enc)
        self._chunks_num = chunks_num

    @abstractmethod
    async def read_chunk(self, no: int = None) -> (int, bytes):
        """Read next chunk

        :param no: int (optional) chunk offset, None if reading from current offset
        :raises StreamEOF if end of stream

        :return (new-chunk-offset, data)
        """
        raise NotImplemented

    @abstractmethod
    async def eof(self) -> bool:
        raise NotImplemented

    async def read(self) -> bytes:
        raw = b''
        async for chunk in self.read_chunked():
            raw += chunk
        return raw

    async def read_chunked(self, src: "AbstractReadOnlyStream" = None):
        if src is None:
            await self.seek_to_chunk(0)
            while not await self.eof():
                _, raw = await self.read_chunk()
                yield raw
        else:
            await src.seek_to_chunk(0)
            while not await src.eof():
                _, chunk = await src.read_chunk()
                raw = await self.decrypt(chunk)
                yield raw


class AbstractWriteOnlyStream(AbstractStream):
    """Stream abstraction for reading operations:
      -----------------------------------------------------------------------
      !!! Warning !!! Supposed chunk-write call has ATOMIC nature and persistent:
        expected Write stream should operate in non-buffering mode OR/AND flush every chunk.
      -----------------------------------------------------------------------

      - cloud storage
      - file-system
      - external storages
      - etc
    """

    def __init__(self, path: str, chunk_size: int = 1024, enc: Optional[BaseStreamEncryption] = None):
        """
        :param chunk_size: size (in bytes) of chunks that stream was splitted to
          !!! actual chunks-sizes may be different (when stream is encoded for example) !!!
          Chunks allow:
            - partially upload/download big data files/streams (control progress)
            - encrypt/decrypt big data partially
            - adv. services like upload/download with pause/resume (for cloud providers for example)
        :param enc: allow encrypt stream chunks
        """
        super().__init__(path, enc)
        self.chunk_size = chunk_size

    @property
    def chunk_size(self) -> int:
        return self._chunk_size

    @chunk_size.setter
    def chunk_size(self, value: int):
        if value <= 0:
            raise StreamInitializationError('Chunk Size must to be > 0 !')
        self._chunk_size = value

    @abstractmethod
    async def write_chunk(self, chunk: bytes, no: int = None) -> (int, int):
        """Write new chunk to stream

        :param chunk: data bytes
        :param no: int (optional) chunk offset, None if writing to end of stream
        :return: (new-chunk-offset, writen data size)
        """
        raise NotImplemented

    async def write(self, data: bytes):
        stream = io.BytesIO(data)
        stream.seek(0)
        while True:
            try:
                chunk = stream.read(self.chunk_size)
                if len(chunk) > 0:
                    await self.write_chunk(chunk)
                else:
                    return
            except EOFError:
                return

    @abstractmethod
    async def truncate(self, no: int = 0):
        """Truncate stream content

        :param no: new stream size will bi limited chunk_no
        """
        raise NotImplemented

    async def copy(self, src: AbstractReadOnlyStream):
        if not src.is_open:
            raise StreamInitializationError('Source stream is closed!')
        async for chunk in src.read_chunked():
            await self.write_chunk(chunk)
            if await src.eof():
                return


class ReadOnlyStreamDecodingWrapper(AbstractReadOnlyStream):

    def __init__(self, src: AbstractReadOnlyStream, enc: BaseStreamEncryption):
        if enc is None:
            raise RuntimeError('You should setup encoding')
        super().__init__(path=src.path, chunks_num=src.chunks_num, enc=enc)
        self.__src = src

    @property
    def is_open(self) -> bool:
        return self.__src.is_open

    @property
    def seekable(self) -> Optional[bool]:
        return self.__src.seekable

    @property
    def chunks_num(self) -> int:
        return self.__src.chunks_num

    @property
    def current_chunk(self) -> int:
        return self.__src.current_chunk

    async def read_chunk(self, no: int = None) -> (int, bytes):
        no, chunk = await self.__src.read_chunk(no)
        if self.enc is not None:
            decoded = await self.decrypt(chunk)
        else:
            decoded = chunk
        return no, decoded

    async def eof(self) -> bool:
        return await self.__src.eof()

    async def open(self):
        await self.__src.open()

    async def close(self):
        await self.__src.close()

    async def seek_to_chunk(self, no: int) -> int:
        no = await self.__src.seek_to_chunk(no)
        return no


class WriteOnlyStreamEncodingWrapper(AbstractWriteOnlyStream):

    def __init__(self, dest: AbstractWriteOnlyStream, enc: BaseStreamEncryption):
        if enc is None:
            raise RuntimeError('You should setup encoding')
        super().__init__(path=dest.path, chunk_size=dest.chunk_size, enc=enc)
        self.__src = dest

    @property
    def is_open(self) -> bool:
        return self.__src.is_open

    @property
    def seekable(self) -> Optional[bool]:
        return self.__src.seekable

    @property
    def chunks_num(self) -> int:
        return self.__src.chunks_num

    @property
    def current_chunk(self) -> int:
        return self.__src.current_chunk

    async def write_chunk(self, chunk: bytes, no: int = None) -> (int, int):
        if self.enc is not None:
            encoded = await self.encrypt(chunk)
        else:
            encoded = chunk
        no, sz = await self.__src.write_chunk(encoded, no)
        return no, sz

    async def truncate(self, no: int = 0):
        await self.__src.truncate(no)

    async def open(self):
        await self.__src.open()

    async def close(self):
        await self.__src.close()

    async def seek_to_chunk(self, no: int) -> int:
        no = await self.__src.seek_to_chunk(no)
        return no



import asyncio
from dataclasses import dataclass
from asyncio.queues import Queue as AsyncQueue
from typing import Dict
from threading import Lock

import sirius_sdk
from sirius_sdk.errors.exceptions import SiriusTimeoutIO, OperationAbortedManually
from sirius_sdk.abstract.api import *
from sirius_sdk.messaging import restore_message_instance
from sirius_sdk.abstract.bus import AbstractBus


@dataclass
class Event:
    topic: str
    payload: bytes
    abort: bool = False


class Observer:

    def __init__(self, client_id: str, queue: AsyncQueue, loop: asyncio.AbstractEventLoop):
        self.__queue = queue
        self.__loop = loop
        self.__client_id = client_id

    @property
    def client_id(self) -> str:
        return self.__client_id

    def notify(self, event: Any):
        coro = self.__queue.put(event)
        # Thread-Safe!!!
        asyncio.run_coroutine_threadsafe(coro, loop=self.__loop)


class Subscriptions:

    def __init__(self):
        self.__lock = Lock()
        # Map topic to observers-list
        self.__subscriptions_singleton: Dict[str, List[Observer]] = {}

    def subscribe(self, topic: str, o: Observer):
        self.__acquire()
        try:
            observers = self.__subscriptions_singleton.get(topic, [])
            if not list(filter(lambda x: x.client_id == o.client_id, observers)):
                observers.append(o)
                self.__subscriptions_singleton[topic] = observers
        finally:
            self.__release()

    def unsubscribe(self, client_id: str, topics: List[str] = None):
        self.__acquire()
        try:
            if topics is not None:
                topics_to_process = topics or []
            else:
                # unsubscribe all topics for specified client_id
                topics_to_process = []
                for topic in self.__subscriptions_singleton.keys():
                    observers = self.__subscriptions_singleton.get(topic, [])
                    if list(filter(lambda x: x.client_id == client_id, observers)):
                        topics_to_process.append(topic)
            # Process
            for topic in topics_to_process:
                observers = self.__subscriptions_singleton.get(topic, [])
                self.__subscriptions_singleton[topic] = [o for o in observers if o.client_id != client_id]
        finally:
            self.__release()

    def notify(self, topic: str, payload: bytes) -> int:
        self.__acquire()
        try:
            observers = self.__subscriptions_singleton.get(topic, [])
        finally:
            self.__release()
        # Notify all
        for o in observers:
            o.notify(event=Event(topic, payload))
        return len(observers)

    def notify_abort(self, client_id: str):
        o: Optional[Observer] = None
        self.__acquire()
        try:
            for topic in self.__subscriptions_singleton.keys():
                observers = self.__subscriptions_singleton.get(topic, [])
                found = list(filter(lambda x: x.client_id == client_id, observers))
                if found:
                    o = found[0]
                    break
        finally:
            self.__release()
        # Fire!!!
        if o is not None:
            o.notify(event=Event(topic='*', payload=b'', abort=True))

    def __acquire(self):
        self.__lock.acquire(blocking=True)

    def __release(self):
        self.__lock.release()


class InMemoryBus(AbstractBus):

    __subscriptions_singleton = Subscriptions()

    def __init__(self, crypto: APICrypto = None, loop: asyncio.AbstractEventLoop = None):
        self.__crypto = crypto or sirius_sdk.Crypto
        self.__queue = AsyncQueue()
        self.__client_id = str(id(self))
        if loop is None:
            loop = asyncio.get_event_loop()
        self.__observer = Observer(client_id=self.__client_id, queue=self.__queue, loop=loop)

    async def subscribe(self, thid: str) -> bool:
        self.__subscriptions_singleton.subscribe(thid, self.__observer)
        return True

    async def subscribe_ext(
            self, sender_vk: List[str], recipient_vk: List[str], protocols: List[str]
    ) -> (bool, List[str]):
        raise NotImplemented

    async def unsubscribe(self, thid: str):
        self.__subscriptions_singleton.unsubscribe(client_id=self.__client_id, topics=[thid])

    async def unsubscribe_ext(self, thids: List[str]):
        self.__subscriptions_singleton.unsubscribe(client_id=self.__client_id, topics=thids)

    async def publish(self, thid: str, payload: bytes) -> int:
        num = self.__subscriptions_singleton.notify(thid, payload)
        return num

    async def get_event(self, timeout: float = None) -> AbstractBus.BytesEvent:
        if timeout is None:
            event: Event = await self.__queue.get()
        else:
            coro = self.__queue.get()
            done, pending = await asyncio.wait([coro], timeout=timeout)
            if done:
                event: Event = list(done)[0].result()
            else:
                raise SiriusTimeoutIO
        #
        if event.abort:
            raise OperationAbortedManually
        else:
            return AbstractBus.BytesEvent(thread_id=event.topic, payload=event.payload)

    async def get_message(self, timeout: float = None) -> AbstractBus.MessageEvent:
        event = await self.get_event(timeout)
        decrypted = await self.__crypto.unpack_message(event.payload)
        ok, msg = restore_message_instance(decrypted['message'])
        if not ok:
            msg = Message(**decrypted['message'])
        return AbstractBus.MessageEvent(
            thread_id=event.thread_id,
            message=msg,
            sender_verkey=decrypted.get('sender_verkey', None),
            recipient_verkey=decrypted.get('recipient_verkey', None)
        )

    async def abort(self):
        self.__subscriptions_singleton.notify_abort(self.__client_id)



from typing import Optional, List, Union, Any

from sirius_sdk.agent.pairwise import AbstractPairwiseList
from sirius_sdk.abstract.api import APICrypto
from sirius_sdk.agent.wallet.abstract.cache import AbstractCache
from sirius_sdk.agent.wallet.abstract.did import AbstractDID
from sirius_sdk.agent.wallet.abstract.anoncreds import AbstractAnonCreds
from sirius_sdk.agent.wallet.abstract.non_secrets import AbstractNonSecrets
from sirius_sdk.abstract.p2p import Pairwise
from sirius_sdk.agent.microledgers.abstract import AbstractMicroledgerList, LedgerMeta, Transaction, \
    AbstractMicroledger, AbstractBatchedAPI
from sirius_sdk.agent.wallet import PurgeOptions, CacheOptions, RetrieveRecordOptions
from sirius_sdk.agent.wallet.abstract import AnonCredSchema

from .core import _current_hub


class DIDProxy(AbstractDID):

    async def create_and_store_my_did(self, did: str = None, seed: str = None, cid: bool = None) -> (str, str):
        service = await _current_hub().get_did()
        return await service.create_and_store_my_did(
            did=did, seed=seed, cid=cid
        )

    async def store_their_did(self, did: str, verkey: str = None) -> None:
        service = await _current_hub().get_did()
        return await service.store_their_did(
            did=did, verkey=verkey
        )

    async def set_did_metadata(self, did: str, metadata: dict = None) -> None:
        service = await _current_hub().get_did()
        return await service.set_did_metadata(
            did=did, metadata=metadata
        )

    async def list_my_dids_with_meta(self) -> List[Any]:
        service = await _current_hub().get_did()
        return await service.list_my_dids_with_meta()

    async def get_did_metadata(self, did) -> Optional[dict]:
        service = await _current_hub().get_did()
        return await service.get_did_metadata(did=did)

    async def key_for_local_did(self, did: str) -> str:
        service = await _current_hub().get_did()
        return await service.key_for_local_did(did=did)

    async def key_for_did(self, pool_name: str, did: str) -> str:
        service = await _current_hub().get_did()
        return await service.key_for_did(pool_name=pool_name, did=did)

    async def create_key(self, seed: str = None) -> str:
        service = await _current_hub().get_did()
        return await service.create_key(seed=seed)

    async def replace_keys_start(self, did: str, seed: str = None) -> str:
        service = await _current_hub().get_did()
        return await service.replace_keys_start(did=did, seed=seed)

    async def replace_keys_apply(self, did: str) -> None:
        service = await _current_hub().get_did()
        await service.replace_keys_apply(did=did)

    async def set_key_metadata(self, verkey: str, metadata: dict) -> None:
        service = await _current_hub().get_did()
        await service.set_key_metadata(verkey=verkey, metadata=metadata)

    async def get_key_metadata(self, verkey: str) -> dict:
        service = await _current_hub().get_did()
        return await service.get_key_metadata(verkey=verkey)

    async def set_endpoint_for_did(self, did: str, address: str, transport_key: str) -> None:
        service = await _current_hub().get_did()
        await service.set_endpoint_for_did(did=did, address=address, transport_key=transport_key)

    async def get_endpoint_for_did(self, pool_name: str, did: str) -> (str, Optional[str]):
        service = await _current_hub().get_did()
        return await service.get_endpoint_for_did(pool_name=pool_name, did=did)

    async def get_my_did_with_meta(self, did: str) -> Any:
        service = await _current_hub().get_did()
        return await service.get_my_did_with_meta(did=did)

    async def abbreviate_verkey(self, did: str, full_verkey: str) -> str:
        service = await _current_hub().get_did()
        return await service.abbreviate_verkey(did=did, full_verkey=full_verkey)

    async def qualify_did(self, did: str, method: str) -> str:
        service = await _current_hub().get_did()
        return await service.qualify_did(did=did, method=method)


class AnonCredsProxy(AbstractAnonCreds):

    async def issuer_create_schema(
            self, issuer_did: str, name: str, version: str, attrs: List[str]
    ) -> (str, AnonCredSchema):
        service = await _current_hub().get_anoncreds()
        return await service.issuer_create_schema(
            issuer_did=issuer_did, name=name,
            version=version, attrs=attrs
        )

    async def issuer_create_and_store_credential_def(
            self, issuer_did: str, schema: dict, tag: str, signature_type: str = None, config: dict = None
    ) -> (str, dict):
        service = await _current_hub().get_anoncreds()
        return await service.issuer_create_and_store_credential_def(
            issuer_did=issuer_did, schema=schema, tag=tag, signature_type=signature_type, config=config
        )

    async def issuer_rotate_credential_def_start(self, cred_def_id: str, config: dict = None) -> dict:
        service = await _current_hub().get_anoncreds()
        return await service.issuer_rotate_credential_def_start(
            cred_def_id=cred_def_id, config=config
        )

    async def issuer_rotate_credential_def_apply(self, cred_def_id: str):
        service = await _current_hub().get_anoncreds()
        await service.issuer_rotate_credential_def_apply(
            cred_def_id=cred_def_id
        )

    async def issuer_create_and_store_revoc_reg(
            self, issuer_did: str, revoc_def_type: Optional[str],
            tag: str, cred_def_id: str, config: dict, tails_writer_handle: int
    ) -> (str, dict, dict):
        service = await _current_hub().get_anoncreds()
        return await service.issuer_create_and_store_revoc_reg(
            issuer_did=issuer_did, revoc_def_type=revoc_def_type,
            tag=tag, cred_def_id=cred_def_id, config=config, tails_writer_handle=tails_writer_handle
        )

    async def issuer_create_credential_offer(self, cred_def_id: str) -> dict:
        service = await _current_hub().get_anoncreds()
        return await service.issuer_create_credential_offer(cred_def_id=cred_def_id)

    async def issuer_create_credential(
            self, cred_offer: dict, cred_req: dict, cred_values: dict,
            rev_reg_id: str = None, blob_storage_reader_handle: int = None
    ) -> (dict, Optional[str], Optional[dict]):
        service = await _current_hub().get_anoncreds()
        return await service.issuer_create_credential(
            cred_offer=cred_offer, cred_req=cred_req, cred_values=cred_values,
            rev_reg_id=rev_reg_id, blob_storage_reader_handle=blob_storage_reader_handle
        )

    async def issuer_revoke_credential(
            self, blob_storage_reader_handle: int, rev_reg_id: str, cred_revoc_id: str
    ) -> dict:
        service = await _current_hub().get_anoncreds()
        return await service.issuer_revoke_credential(
            blob_storage_reader_handle=blob_storage_reader_handle, rev_reg_id=rev_reg_id, cred_revoc_id=cred_revoc_id
        )

    async def issuer_merge_revocation_registry_deltas(self, rev_reg_delta: dict, other_rev_reg_delta: dict) -> dict:
        service = await _current_hub().get_anoncreds()
        return await service.issuer_merge_revocation_registry_deltas(
            rev_reg_delta=rev_reg_delta, other_rev_reg_delta=other_rev_reg_delta
        )

    async def prover_create_master_secret(self, master_secret_name: str = None) -> str:
        service = await _current_hub().get_anoncreds()
        return await service.prover_create_master_secret(
            master_secret_name=master_secret_name
        )

    async def prover_create_credential_req(
            self, prover_did: str, cred_offer: dict, cred_def: dict, master_secret_id: str
    ) -> (dict, dict):
        service = await _current_hub().get_anoncreds()
        return await service.prover_create_credential_req(
            prover_did=prover_did, cred_offer=cred_offer, cred_def=cred_def, master_secret_id=master_secret_id
        )

    async def prover_set_credential_attr_tag_policy(
            self, cred_def_id: str, tag_attrs: Optional[dict], retroactive: bool
    ) -> None:
        service = await _current_hub().get_anoncreds()
        await service.prover_set_credential_attr_tag_policy(
            cred_def_id=cred_def_id, tag_attrs=tag_attrs, retroactive=retroactive
        )

    async def prover_get_credential_attr_tag_policy(self, cred_def_id: str) -> dict:
        service = await _current_hub().get_anoncreds()
        return await service.prover_get_credential_attr_tag_policy(
            cred_def_id=cred_def_id
        )

    async def prover_store_credential(
            self, cred_id: Optional[str], cred_req_metadata: dict, cred: dict, cred_def: dict, rev_reg_def: dict = None
    ) -> str:
        service = await _current_hub().get_anoncreds()
        return await service.prover_store_credential(
            cred_id=cred_id, cred_req_metadata=cred_req_metadata, cred=cred,
            cred_def=cred_def, rev_reg_def=rev_reg_def
        )

    async def prover_get_credential(self, cred_id: str) -> dict:
        service = await _current_hub().get_anoncreds()
        return await service.prover_get_credential(cred_id=cred_id)

    async def prover_delete_credential(self, cred_id: str) -> None:
        service = await _current_hub().get_anoncreds()
        await service.prover_delete_credential(cred_id=cred_id)

    async def prover_get_credentials(self, filters: dict) -> List[dict]:
        service = await _current_hub().get_anoncreds()
        return await service.prover_get_credentials(filters=filters)

    async def prover_search_credentials(self, query: dict) -> List[dict]:
        service = await _current_hub().get_anoncreds()
        return await service.prover_search_credentials(query=query)

    async def prover_get_credentials_for_proof_req(self, proof_request: dict) -> dict:
        service = await _current_hub().get_anoncreds()
        return await service.prover_get_credentials_for_proof_req(proof_request=proof_request)

    async def prover_search_credentials_for_proof_req(
            self, proof_request: dict, extra_query: dict = None, limit_referents: int = 1
    ) -> dict:
        service = await _current_hub().get_anoncreds()
        return await service.prover_search_credentials_for_proof_req(
            proof_request=proof_request, extra_query=extra_query, limit_referents=limit_referents
        )

    async def prover_create_proof(
            self, proof_req: dict, requested_credentials: dict,
            master_secret_name: str, schemas: dict, credential_defs: dict, rev_states: dict
    ) -> dict:
        service = await _current_hub().get_anoncreds()
        return await service.prover_create_proof(
            proof_req=proof_req, requested_credentials=requested_credentials,
            master_secret_name=master_secret_name, schemas=schemas,
            credential_defs=credential_defs, rev_states=rev_states
        )

    async def verifier_verify_proof(
            self, proof_request: dict, proof: dict, schemas: dict,
            credential_defs: dict, rev_reg_defs: dict, rev_regs: dict
    ) -> bool:
        service = await _current_hub().get_anoncreds()
        return await service.verifier_verify_proof(
            proof_request=proof_request, proof=proof,
            schemas=schemas, credential_defs=credential_defs,
            rev_reg_defs=rev_reg_defs, rev_regs=rev_regs
        )

    async def create_revocation_state(
            self, blob_storage_reader_handle: int, rev_reg_def: dict,
            rev_reg_delta: dict, timestamp: int, cred_rev_id: str
    ) -> dict:
        service = await _current_hub().get_anoncreds()
        return await service.create_revocation_state(
            blob_storage_reader_handle=blob_storage_reader_handle, rev_reg_def=rev_reg_def,
            rev_reg_delta=rev_reg_delta, timestamp=timestamp, cred_rev_id=cred_rev_id
        )

    async def update_revocation_state(
            self, blob_storage_reader_handle: int, rev_state: dict,
            rev_reg_def: dict, rev_reg_delta: dict, timestamp: int, cred_rev_id: str
    ) -> dict:
        service = await _current_hub().get_anoncreds()
        return await service.update_revocation_state(
            blob_storage_reader_handle=blob_storage_reader_handle, rev_state=rev_state,
            rev_reg_def=rev_reg_def, rev_reg_delta=rev_reg_delta, timestamp=timestamp, cred_rev_id=cred_rev_id
        )

    async def generate_nonce(self) -> str:
        service = await _current_hub().get_anoncreds()
        return await service.generate_nonce()

    async def to_unqualified(self, entity: str) -> str:
        service = await _current_hub().get_anoncreds()
        return await service.to_unqualified(entity=entity)


class CryptoProxy(APICrypto):

    async def create_key(self, seed: str = None, crypto_type: str = None) -> str:
        service = await _current_hub().get_crypto()
        return await service.create_key(seed=seed, crypto_type=crypto_type)

    async def set_key_metadata(self, verkey: str, metadata: dict) -> None:
        service = await _current_hub().get_crypto()
        return await service.set_key_metadata(verkey=verkey, metadata=metadata)

    async def get_key_metadata(self, verkey: str) -> Optional[dict]:
        service = await _current_hub().get_crypto()
        return await service.get_key_metadata(verkey=verkey)

    async def crypto_sign(self, signer_vk: str, msg: bytes) -> bytes:
        service = await _current_hub().get_crypto()
        return await service.crypto_sign(signer_vk=signer_vk, msg=msg)

    async def crypto_verify(self, signer_vk: str, msg: bytes, signature: bytes) -> bool:
        service = await _current_hub().get_crypto()
        return await service.crypto_verify(signer_vk=signer_vk, msg=msg, signature=signature)

    async def anon_crypt(self, recipient_vk: str, msg: bytes) -> bytes:
        service = await _current_hub().get_crypto()
        return await service.anon_crypt(recipient_vk=recipient_vk, msg=msg)

    async def anon_decrypt(self, recipient_vk: str, encrypted_msg: bytes) -> bytes:
        service = await _current_hub().get_crypto()
        return await service.anon_decrypt(recipient_vk=recipient_vk, encrypted_msg=encrypted_msg)

    async def pack_message(self, message: Any, recipient_verkeys: list, sender_verkey: str = None) -> bytes:
        service = await _current_hub().get_crypto()
        return await service.pack_message(
            message=message, recipient_verkeys=recipient_verkeys, sender_verkey=sender_verkey
        )

    async def unpack_message(self, jwe: bytes) -> dict:
        service = await _current_hub().get_crypto()
        return await service.unpack_message(jwe=jwe)


class MicroledgersProxy(AbstractMicroledgerList):

    async def batched(self) -> AbstractBatchedAPI:
        service = await _current_hub().get_microledgers()
        batched = await service.batched()
        return batched

    async def create(
            self, name: str, genesis: Union[List[Transaction], List[dict]]
    ) -> (AbstractMicroledger, List[Transaction]):
        service = await _current_hub().get_microledgers()
        return await service.create(name, genesis)

    async def ledger(self, name: str) -> AbstractMicroledger:
        service = await _current_hub().get_microledgers()
        return await service.ledger(name)

    async def reset(self, name: str):
        service = await _current_hub().get_microledgers()
        await service.reset(name)

    async def is_exists(self, name: str):
        service = await _current_hub().get_microledgers()
        return await service.is_exists(name)

    async def leaf_hash(self, txn: Union[Transaction, bytes]) -> bytes:
        service = await _current_hub().get_microledgers()
        return await service.leaf_hash(txn)

    async def list(self) -> List[LedgerMeta]:
        service = await _current_hub().get_microledgers()
        return await service.list()

    async def acquire(self, names: List[str], lock_timeout: float):
        service = await _current_hub().get_microledgers()
        return await service.acquire(names, lock_timeout)

    async def release(self):
        service = await _current_hub().get_microledgers()
        await service.release()


class PairwiseProxy(AbstractPairwiseList):

    async def create(self, pairwise: Pairwise):
        service = await _current_hub().get_pairwise_list()
        await service.create(pairwise)

    async def update(self, pairwise: Pairwise):
        service = await _current_hub().get_pairwise_list()
        await service.update(pairwise)

    async def is_exists(self, their_did: str) -> bool:
        service = await _current_hub().get_pairwise_list()
        return await service.is_exists(their_did)

    async def ensure_exists(self, pairwise: Pairwise):
        service = await _current_hub().get_pairwise_list()
        await service.ensure_exists(pairwise)

    async def load_for_did(self, their_did: str) -> Optional[Pairwise]:
        service = await _current_hub().get_pairwise_list()
        return await service.load_for_did(their_did)

    async def load_for_verkey(self, their_verkey: str) -> Optional[Pairwise]:
        service = await _current_hub().get_pairwise_list()
        return await service.load_for_verkey(their_verkey)

    async def _start_loading(self):
        service = await _current_hub().get_pairwise_list()
        await service._start_loading()

    async def _partial_load(self) -> (bool, List[Pairwise]):
        service = await _current_hub().get_pairwise_list()
        return await service._partial_load()

    async def _stop_loading(self):
        service = await _current_hub().get_pairwise_list()
        await service._stop_loading()


class CacheProxy(AbstractCache):

    async def get_schema(self, pool_name: str, submitter_did: str, id_: str, options: CacheOptions) -> dict:
        service = await _current_hub().get_cache()
        return await service.get_schema(
            pool_name=pool_name, submitter_did=submitter_did, id_=id_, options=options
        )

    async def get_cred_def(self, pool_name: str, submitter_did: str, id_: str, options: CacheOptions) -> dict:
        service = await _current_hub().get_cache()
        return await service.get_cred_def(
            pool_name=pool_name,
            submitter_did=submitter_did,
            id_=id_, options=options
        )

    async def purge_schema_cache(self, options: PurgeOptions) -> None:
        service = await _current_hub().get_cache()
        await service.purge_schema_cache(options=options)

    async def purge_cred_def_cache(self, options: PurgeOptions) -> None:
        service = await _current_hub().get_cache()
        await service.purge_cred_def_cache(options=options)


class NonSecretsProxy(AbstractNonSecrets):

    async def add_wallet_record(self, type_: str, id_: str, value: str, tags: dict = None) -> None:
        service = await _current_hub().get_non_secrets()
        return await service.add_wallet_record(type_=type_, id_=id_, value=value, tags=tags)

    async def update_wallet_record_value(self, type_: str, id_: str, value: str) -> None:
        service = await _current_hub().get_non_secrets()
        return await service.update_wallet_record_value(type_=type_, id_=id_, value=value)

    async def update_wallet_record_tags(self, type_: str, id_: str, tags: dict) -> None:
        service = await _current_hub().get_non_secrets()
        return await service.update_wallet_record_tags(type_=type_, id_=id_, tags=tags)

    async def add_wallet_record_tags(self, type_: str, id_: str, tags: dict) -> None:
        service = await _current_hub().get_non_secrets()
        return await service.add_wallet_record_tags(type_=type_, id_=id_, tags=tags)

    async def delete_wallet_record_tags(self, type_: str, id_: str, tag_names: List[str]) -> None:
        service = await _current_hub().get_non_secrets()
        return await service.delete_wallet_record_tags(type_=type_, id_=id_, tag_names=tag_names)

    async def delete_wallet_record(self, type_: str, id_: str) -> None:
        service = await _current_hub().get_non_secrets()
        return await service.delete_wallet_record(type_=type_, id_=id_)

    async def get_wallet_record(self, type_: str, id_: str, options: RetrieveRecordOptions) -> Optional[dict]:
        service = await _current_hub().get_non_secrets()
        return await service.get_wallet_record(type_=type_, id_=id_, options=options)

    async def wallet_search(self, type_: str, query: dict, options: RetrieveRecordOptions, limit: int = 1) -> (List[dict], int):
        service = await _current_hub().get_non_secrets()
        return await service.wallet_search(type_=type_, query=query, options=options, limit=limit)



import asyncio

import pytest

import sirius_sdk
from sirius_sdk.hub import _current_hub

from tests.helpers import ServerTestSuite


@pytest.mark.asyncio
async def test_sane(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent1')
    async with sirius_sdk.context(params['server_address'], params['credentials'], params['p2p']):
        inst1 = _current_hub()
        inst2 = _current_hub()
    assert id(inst1) == id(inst2)

    params1 = test_suite.get_agent_params('agent1')
    params2 = test_suite.get_agent_params('agent2')

    async with sirius_sdk.context(params1['server_address'], params1['credentials'], params1['p2p']):
        ping1 = await sirius_sdk.ping()
        endpoints1 = await sirius_sdk.endpoints()
        my_did_list1 = await sirius_sdk.DID.list_my_dids_with_meta()

    async with sirius_sdk.context(params2['server_address'], params2['credentials'], params2['p2p']):
        ping2 = await sirius_sdk.ping()
        endpoints2 = await sirius_sdk.endpoints()
        my_did_list2 = await sirius_sdk.DID.list_my_dids_with_meta()

    endpoints1 = [e.address for e in endpoints1]
    endpoints2 = [e.address for e in endpoints2]
    my_did_list1 = [d['did'] for d in my_did_list1]
    my_did_list2 = [d['did'] for d in my_did_list2]
    assert ping1 is True
    assert ping2 is True
    assert set(endpoints1) != set(endpoints2)
    assert set(my_did_list1) != set(my_did_list2)


@pytest.mark.asyncio
async def test_aborting(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent1')
    agent1 = None
    agent2 = None
    async with sirius_sdk.context(params['server_address'], params['credentials'], params['p2p']):
        hub = _current_hub()
        async with hub.get_agent_connection_lazy() as agent:
            agent1 = agent
            ok1 = await agent1.ping()
            assert ok1 is True

        await hub.abort()

        async with hub.get_agent_connection_lazy() as agent:
            agent2 = agent
            ok2 = await agent2.ping()
            assert ok2 is True

    assert id(agent1) != id(agent2)
    assert agent1 is not None
    assert agent2 is not None


@pytest.mark.asyncio
async def test_spawn_child_routines(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent1')

    hub_parent = None
    hub_child1 = None
    hub_child2 = None

    async def child1():
        nonlocal hub_child1
        hub_child1 = _current_hub()

    async def child2():
        nonlocal hub_child2
        hub_child2 = _current_hub()

    async def parent():
        nonlocal hub_parent
        hub_parent = _current_hub()
        await child1()
        fut = asyncio.ensure_future(child2())
        await asyncio.wait([fut])

    async with sirius_sdk.context(params['server_address'], params['credentials'], params['p2p']):
        await parent()

    assert id(hub_parent) == id(hub_child1)
    assert id(hub_parent) != id(hub_child2)
    assert id(hub_child1) != id(hub_child2)



import json
import base64
from typing import Any, Optional

from sirius_sdk.abstract.api import APICrypto
from sirius_sdk.abstract.storage import AbstractKeyValueStorage
from sirius_sdk.encryption import ed25519
from sirius_sdk.encryption import b58_to_bytes
from sirius_sdk.errors.exceptions import SiriusCryptoError
from sirius_sdk.hub.defaults.default_storage import InMemoryKeyValueStorage


class DefaultCryptoService(APICrypto):

    DB_KEYS = 'crypto_keys'
    DB_META = 'crypto_metadata'

    def __init__(self, storage: AbstractKeyValueStorage = None):
        self.storage = storage or InMemoryKeyValueStorage()

    async def create_key(self, seed: str = None, crypto_type: str = None) -> str:
        pk_bytes, sk_bytes = ed25519.create_keypair(seed.encode() if seed else None)
        pk_b58 = ed25519.bytes_to_b58(pk_bytes)
        await self.storage.select_db(self.DB_KEYS)
        await self.storage.set(pk_b58, sk_bytes)
        return pk_b58

    async def set_key_metadata(self, verkey: str, metadata: dict) -> None:
        await self.__check_verkey_exists(verkey)
        await self.storage.select_db(self.DB_META)
        await self.storage.set(verkey, metadata)

    async def get_key_metadata(self, verkey: str) -> Optional[dict]:
        await self.__check_verkey_exists(verkey)
        await self.storage.select_db(self.DB_META)
        meta = await self.storage.get(verkey)
        return meta

    async def crypto_sign(self, signer_vk: str, msg: bytes) -> bytes:
        await self.__check_verkey_exists(signer_vk)
        await self.storage.select_db(self.DB_KEYS)
        sk_bytes = await self.storage.get(signer_vk)
        signed = ed25519.sign_message(msg, sk_bytes)
        return signed

    async def crypto_verify(self, signer_vk: str, msg: bytes, signature: bytes) -> bool:
        signer_vk_bytes = b58_to_bytes(signer_vk)
        success = ed25519.verify_signed_message(signer_vk_bytes, msg, signature)
        return success

    async def anon_crypt(self, recipient_vk: str, msg: bytes) -> bytes:
        vk_bytes = b58_to_bytes(recipient_vk)
        encrypted = ed25519.crypto_box_seal(msg, vk_bytes)
        return encrypted

    async def anon_decrypt(self, recipient_vk: str, encrypted_msg: bytes) -> bytes:
        await self.__check_verkey_exists(recipient_vk)
        await self.storage.select_db(self.DB_KEYS)
        sk_bytes = await self.storage.get(recipient_vk)
        vk_bytes = b58_to_bytes(recipient_vk)
        decrypt = ed25519.crypto_box_seal_open(vk_bytes, sk_bytes, encrypted_msg)
        return decrypt

    async def pack_message(self, message: Any, recipient_verkeys: list, sender_verkey: str = None) -> bytes:
        if sender_verkey is not None:
            await self.__check_verkey_exists(sender_verkey)
        recipient_verkeys_bytes = []
        for vk_b58 in recipient_verkeys:
            recipient_verkeys_bytes.append(b58_to_bytes(vk_b58))
        sender_verkey_bytes = b58_to_bytes(sender_verkey)
        if sender_verkey is None:
            sender_sigkey_bytes = None
        else:
            await self.storage.select_db(self.DB_KEYS)
            sender_sigkey_bytes = await self.storage.get(sender_verkey)
        jwe = ed25519.pack_message(
            message=message,
            to_verkeys=recipient_verkeys_bytes,
            from_verkey=sender_verkey_bytes,
            from_sigkey=sender_sigkey_bytes
        )
        return jwe

    async def unpack_message(self, jwe: bytes) -> dict:
        try:
            msg = json.loads(jwe.decode())
        except Exception as e:
            if isinstance(e, json.JSONDecodeError) or isinstance(e, UnicodeError):
                raise SiriusCryptoError('Unexpected packed message format')
            else:
                raise
        if 'protected' in msg:
            try:
                recip = base64.b64decode(msg['protected'])
                recip_json = json.loads(recip.decode())
                recipients = recip_json.get('recipients', [])
                my_vk, my_sk = None, None
                await self.storage.select_db(self.DB_KEYS)
                # Search my_key in headers
                for rcp in recipients:
                    header = rcp.get('header', {})
                    vk = header.get('kid', None)
                    if vk is not None:
                        my_sk = await self.storage.get(vk)
                        if my_sk is not None:
                            my_vk = b58_to_bytes(vk)
                            break

                if not my_sk:
                    raise SiriusCryptoError('Unknown key in recipient list')
                unpacked = ed25519.unpack_message(
                    enc_message=jwe, my_verkey=my_vk, my_sigkey=my_sk
                )
                return {
                    'message': unpacked[0],
                    'recipient_verkey': unpacked[2],
                    'sender_verkey': unpacked[1]
                }
            except ValueError:
                raise SiriusCryptoError('Unexpected packed message format')
        else:
            raise SiriusCryptoError('Unexpected packed message format')

    async def __check_verkey_exists(self, verkey: str):
        await self.storage.select_db(self.DB_KEYS)
        sk = await self.storage.get(verkey)
        if sk is None:
            raise SiriusCryptoError(f'Unknown verkey "{verkey}"')
